from onegov.activity.models.publication_request import PublicationRequest
from onegov.core.collection import GenericCollection


class PublicationRequestCollection(GenericCollection):

    @property
    def model_class(self):
        return PublicationRequest



from onegov.activity.models import Attendee
from onegov.core.collection import GenericCollection


class AttendeeCollection(GenericCollection):

    @property
    def model_class(self):
        return Attendee

    def by_user(self, user):
        return self.query().filter(self.model_class.username == user.username)

    def by_username(self, username):
        return self.query().filter(self.model_class.username == username)

    def add(self, user, name, birth_date, gender, notes=None):

        return super().add(
            username=user.username,
            name=name,
            birth_date=birth_date,
            gender=gender,
            notes=notes
        )



from onegov.activity.collections.activity import ActivityFilter
from onegov.activity.collections.activity import ActivityCollection
from onegov.activity.collections.attendee import AttendeeCollection
from onegov.activity.collections.booking import BookingCollection
from onegov.activity.collections.invoice import InvoiceCollection
from onegov.activity.collections.occasion import OccasionCollection
from onegov.activity.collections.period import PeriodCollection
from onegov.activity.collections.publication_request import \
    PublicationRequestCollection

__all__ = [
    'ActivityCollection',
    'ActivityFilter',
    'AttendeeCollection',
    'BookingCollection',
    'InvoiceCollection',
    'OccasionCollection',
    'PeriodCollection',
    'PublicationRequestCollection'
]



import hashlib

from onegov.activity.models import Activity, Attendee, Booking, Occasion
from onegov.user import User
from sqlalchemy import func


class Scoring(object):
    """ Provides scoring based on a number of criteria.

    A criteria is a callable which takes a booking and returns a score.
    The final score is the sum of all criteria scores.

    """

    def __init__(self, criteria=None):
        self.criteria = criteria or [PreferMotivated()]

    def __call__(self, booking):
        return sum(criterium(booking) for criterium in self.criteria)

    @classmethod
    def from_settings(cls, settings, session):
        scoring = cls()

        # always prefer groups
        scoring.criteria.append(PreferGroups.from_session(session))

        if settings.get('prefer_in_age_bracket'):
            scoring.criteria.append(
                PreferInAgeBracket.from_session(session))

        if settings.get('prefer_organiser'):
            scoring.criteria.append(
                PreferOrganiserChildren.from_session(session))

        if settings.get('prefer_admins'):
            scoring.criteria.append(
                PreferAdminChildren.from_session(session))

        return scoring

    @property
    def settings(self):
        classes = {c.__class__ for c in self.criteria}
        settings = {}

        if PreferInAgeBracket in classes:
            settings['prefer_in_age_bracket'] = True

        if PreferOrganiserChildren in classes:
            settings['prefer_organiser'] = True

        if PreferAdminChildren in classes:
            settings['prefer_admins'] = True

        return settings


class PreferMotivated(object):
    """ Scores "motivated" bookings higher. A motivated booking is simply a
    booking with a higher priority (an attendee would favor a booking he's
    excited about.)

    """

    @classmethod
    def from_session(cls, session):
        return cls()

    def __call__(self, booking):
        return booking.priority


class PreferInAgeBracket(object):
    """ Scores bookings whose attendees fall into the age-bracket of the
    occasion higher.

    If the attendee falls into the age-bracket, the score is 1.0. Each year
    difference results in a penalty of 0.1, until 0.0 is reached.

    """

    def __init__(self, get_age_range, get_attendee_age):
        self.get_age_range = get_age_range
        self.get_attendee_age = get_attendee_age

    @classmethod
    def from_session(cls, session):
        attendees = None
        occasions = None

        def get_age_range(booking):
            nonlocal occasions, session

            if occasions is None:
                occasions = {
                    o.id: o.age
                    for o in session.query(Occasion.id, Occasion.age)
                    .filter(Occasion.period_id == booking.period_id)}

            return (
                occasions[booking.occasion_id].lower,
                occasions[booking.occasion_id].upper - 1
            )

        def get_attendee_age(booking):
            nonlocal attendees, session

            if attendees is None:
                attendees = {a.id: a.age for a in session.query(
                    Attendee.id, Attendee.age)}

            return attendees[booking.attendee_id]

        return cls(get_age_range, get_attendee_age)

    def __call__(self, booking):
        min_age, max_age = self.get_age_range(booking)
        attendee_age = self.get_attendee_age(booking)

        if min_age <= attendee_age and attendee_age <= max_age:
            return 1.0
        else:
            difference = min(
                abs(min_age - attendee_age),
                abs(max_age - attendee_age)
            )
            return 1.0 - min(1.0, difference / 10.0)


class PreferOrganiserChildren(object):
    """ Scores bookings of children higher if their parents are organisers.

    This is basically an incentive to become an organiser. A child whose parent
    is an organiser gets a score of 1.0, if the parent is not an organiser
    a score 0.0 is returned.

    """

    def __init__(self, get_is_organiser_child):
        self.get_is_organiser_child = get_is_organiser_child

    @classmethod
    def from_session(cls, session):
        organisers = None

        def get_is_organiser_child(booking):
            nonlocal organisers

            if organisers is None:
                organisers = {
                    a.username
                    for a in session.query(Activity.username)
                    .filter(Activity.id.in_(
                        session.query(Occasion.activity_id)
                        .filter(Occasion.period_id == booking.period_id)
                        .subquery()
                    ))
                }

            return booking.username in organisers

        return cls(get_is_organiser_child)

    def __call__(self, booking):
        return self.get_is_organiser_child(booking) and 1.0 or 0.0


class PreferAdminChildren(object):
    """ Scores bookings of children higher if their parents are admins. """

    def __init__(self, get_is_association_child):
        self.get_is_association_child = get_is_association_child

    @classmethod
    def from_session(cls, session):
        members = None

        def get_is_association_child(booking):
            nonlocal members

            if members is None:
                members = {
                    u.username for u in session.query(User)
                    .filter(User.role == 'admin')
                    .filter(User.active == True)
                }

            return booking.username in members

        return cls(get_is_association_child)

    def __call__(self, booking):
        return self.get_is_association_child(booking) and 1.0 or 0.0


class PreferGroups(object):
    """ Scores group bookings higher than other bookings. Groups get a boost
    by size:

    - 2 people: 1.0
    - 3 people: 0.8
    - 4 people: 0.6
    - more people: 0.5

    This preference gives an extra boost to unprioritised bookings, to somewhat
    level out bookings in groups that used no star (otherwise a group
    might be split up because someone didn't star the booking).

    Additionally a unique boost between 0.010000000 to 0.099999999 is given to
    each group depending on the group name. This should ensure that competing
    groups generally do not have the same score. So an occasion will generally
    prefer the members of one group over members of another group.

    """

    def __init__(self, get_group_score):
        self.get_group_score = get_group_score

    @classmethod
    def from_session(cls, session):
        group_scores = None

        def unique_score_modifier(group_code):
            digest = hashlib.sha1(group_code.encode('utf-8')).hexdigest()[:8]
            number = int(digest, 16)

            return float('0.0' + str(number)[:8])

        def get_group_score(booking):
            nonlocal group_scores

            if group_scores is None:
                query = session.query(Booking).with_entities(
                    Booking.group_code,
                    func.count(Booking.group_code).label('count')
                ).filter(
                    Booking.group_code != None,
                    Booking.period_id == booking.period_id
                ).group_by(
                    Booking.group_code
                ).having(
                    func.count(Booking.group_code) > 1
                )

                group_scores = {
                    r.group_code:
                    max(.5, 1.0 - 0.2 * (r.count - 2))
                    + unique_score_modifier(r.group_code)

                    for r in query
                }

            return group_scores.get(booking.group_code, 0)

        return get_group_score

    def __call__(self, booking):
        offset = 0 if booking.priority else 1
        return self.get_group_score(booking) + offset



import random

from datetime import date, timedelta
from freezegun import freeze_time
from functools import partial
from onegov.activity.matching import deferred_acceptance_from_database
from onegov.activity.matching import PreferAdminChildren
from onegov.activity.matching import PreferGroups
from onegov.activity.matching import PreferInAgeBracket
from onegov.activity.matching import PreferOrganiserChildren
from onegov.activity.matching import Scoring
from onegov.core.utils import Bunch
from psycopg2.extras import NumericRange
from uuid import uuid4


match = partial(
    deferred_acceptance_from_database,
    stability_check=True,
    validity_check=True,
    hard_budget=True
)


def new_occasion(collections, period, offset, length,
                 activity_name=None, username="owner@example.org",
                 spots=(0, 10), age=(0, 10)):

    activity_name = activity_name or uuid4().hex
    activity = collections.activities.by_name(activity_name)

    if not activity:
        activity = collections.activities.add(activity_name, username=username)

    return collections.occasions.add(
        start=period.prebooking_start + timedelta(days=offset),
        end=period.prebooking_start + timedelta(days=offset + length),
        timezone="Europe/Zurich",
        activity=activity,
        period=period,
        spots=spots,
        age=age
    )


def new_attendee(collections, name=None, birth_date=None, user=None):

    return collections.attendees.add(
        name=name or uuid4().hex,
        birth_date=birth_date or date(2000, 1, 1),
        user=user or Bunch(username="owner@example.org"),
        gender=random.choice(('male', 'female')))


def test_simple_match(session, owner, collections, prebooking_period):
    """ Tests the matching of two occasions on the same activity with no
    conflicting signups.

    """

    o1 = new_occasion(collections, prebooking_period, 0, 1, spots=(0, 1))
    o2 = new_occasion(collections, prebooking_period, 0, 1, spots=(0, 1))

    a1 = new_attendee(collections)
    a2 = new_attendee(collections)

    collections.bookings.add(owner, a1, o1)
    collections.bookings.add(owner, a2, o2)

    # multiple runs lead to the same result
    for i in range(0, 2):
        match(session, prebooking_period.id)

        bookings = collections.bookings.query().all()

        assert len(bookings) == 2

        assert bookings[0].state == 'accepted'
        assert bookings[1].state == 'accepted'

        assert a1.happiness(prebooking_period.id) == 1.0
        assert a2.happiness(prebooking_period.id) == 1.0


def test_changing_priorities(session, owner, collections, prebooking_period):
    """ Makes sure that changing priorities lead to a changing match. """

    # only one available spot
    o = new_occasion(collections, prebooking_period, 0, 1, spots=(0, 1))

    # two attendees
    a1 = new_attendee(collections)
    a2 = new_attendee(collections)

    # one booking is prioritised
    b1 = collections.bookings.add(owner, a1, o)
    b2 = collections.bookings.add(owner, a2, o)

    b1.priority = 1
    b2.priority = 0

    match(session, prebooking_period.id)

    assert b1.state == 'accepted'
    assert b2.state == 'open'

    b1.priority = 0
    b2.priority = 1

    match(session, prebooking_period.id)

    assert b1.state == 'open'
    assert b2.state == 'accepted'


@freeze_time("2016-11-22")
def test_prefer_in_age_bracket(session, owner, collections, prebooking_period):
    o = new_occasion(
        collections, prebooking_period, 0, 1, spots=(0, 1), age=(10, 20))

    # two attendees, one is inside the age bracket, the other is outside
    a1 = new_attendee(collections, birth_date=date(1980, 11, 22))
    a2 = new_attendee(collections, birth_date=date(2000, 11, 22))

    b1 = collections.bookings.add(owner, a1, o)
    b2 = collections.bookings.add(owner, a2, o)

    match(session, prebooking_period.id, score_function=Scoring(
        criteria=[PreferInAgeBracket.from_session(session)]
    ))

    assert b1.state == 'open'
    assert b2.state == 'accepted'

    o.age = NumericRange(20, 50)

    match(session, prebooking_period.id, score_function=Scoring(
        criteria=[PreferInAgeBracket.from_session(session)]
    ))

    assert b1.state == 'accepted'
    assert b2.state == 'open'


def test_prefer_organisers_of_period(session, owner, secondary_owner,
                                     collections, prebooking_period,
                                     inactive_period):

    # if someone was an organiser in another period, it doesn't count
    new_occasion(collections, inactive_period, 0, 1, spots=(0, 1),
                 username=secondary_owner.username, activity_name="old")

    # what counts is being an organiser in the current period
    o = new_occasion(collections, prebooking_period, 0, 1, spots=(0, 1),
                     username=owner.username, activity_name="new")

    # two attendees, one belonging to an organiser
    a1 = new_attendee(collections, user=owner)
    a2 = new_attendee(collections, user=secondary_owner)

    b1 = collections.bookings.add(owner, a1, o)
    b2 = collections.bookings.add(secondary_owner, a2, o)

    match(session, prebooking_period.id, score_function=Scoring(
        criteria=[PreferOrganiserChildren.from_session(session)]
    ))

    assert b1.state == 'accepted'
    assert b2.state == 'open'


def test_prefer_organisers_over_members(session, owner, member,
                                        collections, prebooking_period):

    # organisers > members
    o = new_occasion(collections, prebooking_period, 0, 1, spots=(0, 1),
                     username=owner.username)

    # two attendees, one belonging to an organiser
    a1 = new_attendee(collections, user=member)
    a2 = new_attendee(collections, user=owner)

    b1 = collections.bookings.add(member, a1, o)
    b2 = collections.bookings.add(owner, a2, o)

    match(session, prebooking_period.id, score_function=Scoring(
        criteria=[PreferOrganiserChildren.from_session(session)]
    ))

    assert b1.state == 'open'
    assert b2.state == 'accepted'


def test_prefer_admin_children(session, owner, member, collections,
                               prebooking_period):

    owner.role = 'admin'

    o = new_occasion(collections, prebooking_period, 0, 1, spots=(0, 1))

    # two attendees, one belonging to an association member (editor)
    a1 = new_attendee(collections, user=member)
    a2 = new_attendee(collections, user=owner)

    b1 = collections.bookings.add(owner, a1, o)
    b2 = collections.bookings.add(member, a2, o)

    match(session, prebooking_period.id, score_function=Scoring(
        criteria=[PreferAdminChildren.from_session(session)]
    ))

    assert b1.state == 'accepted'
    assert b2.state == 'open'


def test_interleaved_dates(session, owner, collections, prebooking_period):

    o1 = new_occasion(collections, prebooking_period, 0, 1)
    collections.occasions.add_date(
        o1,
        o1.dates[0].end + timedelta(days=2, seconds=1),
        o1.dates[0].end + timedelta(days=3),
        o1.dates[0].timezone
    )

    o2 = new_occasion(collections, prebooking_period, 2, 1)
    collections.occasions.add_date(
        o2,
        o2.dates[0].end + timedelta(days=3, seconds=1),
        o2.dates[0].end + timedelta(days=4),
        o2.dates[0].timezone
    )

    a1 = new_attendee(collections, user=owner)

    b1 = collections.bookings.add(owner, a1, o1, priority=1)
    b2 = collections.bookings.add(owner, a1, o2, priority=0)

    match(session, prebooking_period.id)

    assert b1.state == 'accepted'
    assert b2.state == 'accepted'


def test_overlapping_dates(session, owner, collections, prebooking_period):

    # only the second dates overlap
    o1 = new_occasion(collections, prebooking_period, 0, 1)
    collections.occasions.add_date(
        o1,
        o1.dates[0].end + timedelta(days=3, seconds=1),
        o1.dates[0].end + timedelta(days=5),
        o1.dates[0].timezone
    )

    o2 = new_occasion(collections, prebooking_period, 2, 1)
    collections.occasions.add_date(
        o2,
        o2.dates[0].end + timedelta(days=2, seconds=1),
        o2.dates[0].end + timedelta(days=3),
        o2.dates[0].timezone
    )

    a1 = new_attendee(collections, user=owner)

    b1 = collections.bookings.add(owner, a1, o1, priority=1)
    b2 = collections.bookings.add(owner, a1, o2, priority=0)

    match(session, prebooking_period.id)

    assert b1.state == 'accepted'
    assert b2.state == 'blocked'


def test_alignment(session, owner, collections, prebooking_period):

    o1 = new_occasion(collections, prebooking_period, 0, 0.5)
    o2 = new_occasion(collections, prebooking_period, 0.5, 0.5)

    a1 = new_attendee(collections, user=owner)

    b1 = collections.bookings.add(owner, a1, o1, priority=1)
    b2 = collections.bookings.add(owner, a1, o2, priority=0)

    match(session, prebooking_period.id)

    assert b1.state == 'accepted'
    assert b2.state == 'accepted'

    prebooking_period.alignment = 'day'

    match(session, prebooking_period.id)
    assert b1.state == 'accepted'
    assert b2.state == 'blocked'


def test_activity_one_occasion(session, owner, collections, prebooking_period):

    o1 = new_occasion(collections, prebooking_period, 0, 1, activity_name='x')
    o2 = new_occasion(collections, prebooking_period, 0, 1, activity_name='x')

    a1 = new_attendee(collections, user=owner)

    b1 = collections.bookings.add(owner, a1, o1, priority=1)
    b2 = collections.bookings.add(owner, a1, o2, priority=0)

    match(session, prebooking_period.id)

    assert b1.state == 'accepted'
    assert b2.state == 'blocked'


def test_prefer_groups(session, owner, collections, prebooking_period):
    o = new_occasion(collections, prebooking_period, 0, 1, spots=(0, 2))

    a1 = new_attendee(collections, user=owner)
    a2 = new_attendee(collections, user=owner)
    a3 = new_attendee(collections, user=owner)

    b1 = collections.bookings.add(owner, a1, o, priority=0)
    b2 = collections.bookings.add(owner, a2, o, priority=0)
    b3 = collections.bookings.add(owner, a3, o, priority=1)

    b1.group_code = b2.group_code = 'foo'

    match(session, prebooking_period.id, score_function=Scoring(
        criteria=[PreferGroups.from_session(session)]
    ))

    assert b1.state == 'accepted'
    assert b2.state == 'accepted'
    assert b3.state == 'open'


def test_prefer_small_groups(session, owner, collections, prebooking_period):
    o = new_occasion(collections, prebooking_period, 0, 1, spots=(0, 2))

    a1 = new_attendee(collections, user=owner)
    a2 = new_attendee(collections, user=owner)
    a3 = new_attendee(collections, user=owner)
    a4 = new_attendee(collections, user=owner)
    a5 = new_attendee(collections, user=owner)

    b1 = collections.bookings.add(owner, a1, o, priority=0)
    b2 = collections.bookings.add(owner, a2, o, priority=0)
    b3 = collections.bookings.add(owner, a3, o, priority=1)
    b4 = collections.bookings.add(owner, a4, o, priority=1)
    b5 = collections.bookings.add(owner, a5, o, priority=1)

    b1.group_code = b2.group_code = 'foo'
    b3.group_code = b4.group_code = b5.group_code = 'bar'

    match(session, prebooking_period.id, score_function=Scoring(
        criteria=[PreferGroups.from_session(session)]
    ))

    assert b1.state == 'accepted'
    assert b2.state == 'accepted'
    assert b3.state == 'open'
    assert b4.state == 'open'
    assert b5.state == 'open'


def test_keep_groups_together(session, owner, collections, prebooking_period):
    # all things being equal, groups are kept together
    o = new_occasion(collections, prebooking_period, 0, 1, spots=(0, 2))

    a1 = new_attendee(collections, user=owner)
    a2 = new_attendee(collections, user=owner)
    a3 = new_attendee(collections, user=owner)
    a4 = new_attendee(collections, user=owner)

    b1 = collections.bookings.add(owner, a1, o, priority=0)
    b2 = collections.bookings.add(owner, a2, o, priority=0)
    b3 = collections.bookings.add(owner, a3, o, priority=0)
    b4 = collections.bookings.add(owner, a4, o, priority=0)

    b1.group_code = b2.group_code = 'foo'
    b3.group_code = b4.group_code = 'bar'

    match(session, prebooking_period.id, score_function=Scoring(
        criteria=[PreferGroups.from_session(session)]
    ))

    assert b1.state == 'accepted'
    assert b2.state == 'accepted'
    assert b3.state == 'open'
    assert b4.state == 'open'

    b1.group_code = b2.group_code = 'bar'
    b3.group_code = b4.group_code = 'foo'

    match(session, prebooking_period.id, score_function=Scoring(
        criteria=[PreferGroups.from_session(session)]
    ))

    assert b1.state == 'open'
    assert b2.state == 'open'
    assert b3.state == 'accepted'
    assert b4.state == 'accepted'


def test_prefer_groups_equal(session, owner, collections, prebooking_period):
    # make sure that given an equal choice between two occasions we prefer
    # the group joining an occasion over joining another as a non-group
    o1 = new_occasion(collections, prebooking_period, 0, 1)
    o2 = new_occasion(collections, prebooking_period, 0, 1)

    a1 = new_attendee(collections, user=owner)
    a2 = new_attendee(collections, user=owner)

    b1 = collections.bookings.add(owner, a1, o1, priority=0)
    b2 = collections.bookings.add(owner, a2, o1, priority=0)
    b3 = collections.bookings.add(owner, a1, o2, priority=0)
    b4 = collections.bookings.add(owner, a2, o2, priority=0)

    b1.group_code = b2.group_code = 'foo'

    match(session, prebooking_period.id, score_function=Scoring(
        criteria=[PreferGroups.from_session(session)]
    ))

    assert b1.state == 'accepted'
    assert b2.state == 'accepted'
    assert b3.state == 'blocked'
    assert b4.state == 'blocked'



import transaction

from cached_property import cached_property
from contextlib import contextmanager
from datetime import datetime, timedelta, time
from faker import Faker
from onegov.activity import Activity
from onegov.activity import ActivityCollection
from onegov.activity import AttendeeCollection
from onegov.activity import BookingCollection
from onegov.activity import OccasionCollection
from onegov.activity import PeriodCollection
from onegov.activity.models import Attendee
from onegov.activity.models import Booking
from onegov.activity.models import Occasion
from onegov.activity.models import OccasionDate
from onegov.activity.models import Period
from onegov.core.utils import normalize_for_url
from onegov.pay import PaymentProviderCollection
from onegov.user import UserCollection
from onegov.user.models import User
from sedate import standardize_date
from sqlalchemy import inspect


class Collections(object):

    def __init__(self, session, activity_type):
        self.session = session
        self.activity_type = activity_type

    @cached_property
    def activities(self):
        return ActivityCollection(self.session, type=self.activity_type)

    @cached_property
    def attendees(self):
        return AttendeeCollection(self.session)

    @cached_property
    def bookings(self):
        return BookingCollection(self.session)

    @cached_property
    def users(self):
        return UserCollection(self.session)

    @cached_property
    def periods(self):
        return PeriodCollection(self.session)

    @cached_property
    def occasions(self):
        return OccasionCollection(self.session)

    @cached_property
    def payment_providers(self):
        return PaymentProviderCollection(self.session)


class Scenario(object):
    """ Helper class to ease the setup of testing fixtures for Feriennet.

    Feriennet has often repetitive steps to get to a test scenario. A period
    needs to be added, an activity, an occasion, an attendee and so on.

    This class offers to do this with the twist that it'll use the order in
    which its methods are being called as extra information.

    So if we add an activity and on the next line add an occasion, the occasion
    will automatically be added to the last created activity.

    This is usually something we might want to be explicit about, but for our
    tests this is helps us cut down on the amount of code needed to setup a
    certain scenario, while making it clearer what we are setting up.

    For example:

        scenario.add_period(confirmed=True)
        scenario.add_activity(title="Learn How to Program")
        scenario.add_occasion()
        scenario.add_attendee()
        scenario.add_booking()

    Here we create a period with an activity in it that has an occasion for
    which an attendee has a booking.

    """

    def __init__(self, session, test_password, activity_model=Activity):
        self.session = session
        self.activity_model = activity_model
        self.activity_type = inspect(activity_model).polymorphic_identity
        self.test_password = test_password

        self.faker = Faker()

        self.activities = []
        self.attendees = []
        self.bookings = []
        self.occasions = []
        self.periods = []
        self.users = []

    def __getattribute__(self, name):
        if name.startswith('add_') and not transaction.manager.manager._txn:
            transaction.begin()

        return super().__getattribute__(name)

    def commit(self):
        transaction.commit()

    @contextmanager
    def update(self):
        self.refresh()
        yield
        self.commit()

    def refresh(self):
        transaction.begin()

        caches = (
            'activities',
            'attendees',
            'bookings',
            'occasions',
            'periods',
            'users'
        )

        for name in caches:
            cache = getattr(self, name)

            for ix, item in enumerate(cache):
                cache[ix] = self.session.merge(item)
                self.session.refresh(cache[ix])

    @cached_property
    def c(self):
        """ Returns useful collections, lazy-loaded. """
        return Collections(self.session, self.activity_type or '*')

    @property
    def latest_activity(self):
        return self.activities and self.activities[-1] or None

    @property
    def latest_booking(self):
        return self.bookings and self.bookings[-1] or None

    @property
    def latest_attendee(self):
        return self.attendees and self.attendees[-1] or None

    @property
    def latest_occasion(self):
        return self.occasions and self.occasions[-1] or None

    @property
    def latest_period(self):
        return self.periods and self.periods[-1] or None

    @property
    def latest_user(self):
        return self.users and self.users[-1] or None

    @property
    def latest_username(self):
        return self.users and self.users[-1].username or None

    @property
    def default_username(self):
        return self.latest_username or 'admin@example.org'

    def date_offset(self, offset):
        return (datetime.now() + timedelta(days=offset)).date()

    def add(self, model, **columns):
        obj = model(**columns)
        self.session.add(obj)

        return obj

    def add_period(self, phase='wishlist', **columns):
        columns.setdefault('title', self.faker.sentence())
        columns.setdefault('active', True)

        if phase == 'wishlist':
            columns.setdefault('prebooking_start', self.date_offset(-1))
            columns.setdefault('prebooking_end', self.date_offset(+1))
            columns.setdefault('execution_start', self.date_offset(+10))
            columns.setdefault('execution_end', self.date_offset(+20))
        elif phase == 'booking':
            columns.setdefault('prebooking_start', self.date_offset(-10))
            columns.setdefault('prebooking_end', self.date_offset(-5))
            columns.setdefault('execution_start', self.date_offset(-1))
            columns.setdefault('execution_end', self.date_offset(+1))
        else:
            raise NotImplementedError

        self.periods.append(self.add(Period, **columns))

        return self.latest_period

    def add_user(self, complete_profile=True, **columns):
        columns.setdefault('role', 'admin')
        columns.setdefault('username', self.faker.email())

        self.users.append(self.add(
            model=User,
            password_hash=self.test_password,
            **columns
        ))

        if complete_profile:
            user = self.latest_user
            user.realname = \
                f'{self.faker.first_name()}\u00A0{self.faker.last_name()}'
            user.data = user.data or {}
            user.data['salutation'] = self.faker.random_element(('mr', 'ms'))
            user.data['address'] = self.faker.address()
            user.data['zip_code'] = self.faker.zipcode()
            user.data['place'] = self.faker.city()
            user.data['political_municipality'] = self.faker.city()
            user.data['emergency'] = f'123 456 789 ({self.faker.name()})'

        return self.latest_user

    def add_activity(self, **columns):
        columns.setdefault('title', self.faker.sentence())
        columns.setdefault('name', normalize_for_url(columns['title']))
        columns.setdefault('username', self.default_username)
        columns.setdefault('state', 'preview')

        self.activities.append(self.add(self.activity_model, **columns))

        return self.latest_activity

    def add_occasion(self, start=None, end=None, dates=None, **columns):
        assert not (dates and start and end)

        columns.setdefault('period', self.latest_period)
        columns.setdefault('activity', self.latest_activity)
        columns.setdefault('age', (0, 20))
        columns.setdefault('spots', (0, 20))

        for key in ('age', 'spots'):
            columns[key] = OccasionCollection.to_half_open_interval(
                *columns[key])

        if not dates and (start and end):
            dates = ((start, end), )

        if not dates:
            period = self.latest_period
            offset = sum(1 for o in self.occasions if o.period == period)

            dates = ((
                datetime.combine(period.execution_start, time()) + timedelta(
                    hours=offset
                ),
                datetime.combine(period.execution_start, time()) + timedelta(
                    hours=offset + 1
                )
            ),)

        self.occasions.append(self.add(model=Occasion, **columns))

        for start, end in dates:
            self.latest_occasion.dates.append(OccasionDate(
                start=standardize_date(start, 'Europe/Zurich'),
                end=standardize_date(end, 'Europe/Zurich'),
                timezone='Europe/Zurich'
            ))

        return self.latest_occasion

    def add_attendee(self, **columns):
        columns.setdefault('name', self.faker.name())
        columns.setdefault('birth_date', self.faker.date_of_birth())
        columns.setdefault('username', self.default_username)
        columns.setdefault('gender', self.faker.random_element((
            'female', 'male'
        )))

        self.attendees.append(self.add(model=Attendee, **columns))

        return self.latest_attendee

    def add_booking(self, **columns):
        columns.setdefault('username', self.default_username)
        columns.setdefault('attendee', self.latest_attendee)
        columns.setdefault('occasion', self.latest_occasion)

        # required by the aggregation performed on the booking
        # (in reality we always create periods, bookings and occasions in
        # separate transactions)
        self.session.flush()
        columns['period_id'] = columns['occasion'].period_id

        self.bookings.append(self.add(model=Booking, **columns))

        return self.latest_booking



import sedate

from datetime import date, datetime
from onegov.activity.models.booking import Booking
from onegov.activity.models.occasion import Occasion
from onegov.core.orm import Base
from onegov.core.orm.mixins import TimestampMixin
from onegov.core.orm.types import UUID, JSON
from sqlalchemy import Boolean
from sqlalchemy import CheckConstraint
from sqlalchemy import column
from sqlalchemy import Column
from sqlalchemy import Date
from sqlalchemy import Index
from sqlalchemy import Integer
from sqlalchemy import Numeric
from sqlalchemy import Text
from sqlalchemy import case, desc, not_, distinct, func
from sqlalchemy.orm import object_session, relationship, joinedload, defer
from sqlalchemy.ext.hybrid import hybrid_property
from uuid import uuid4


class Period(Base, TimestampMixin):

    __tablename__ = 'periods'

    # It's doubtful that the Ferienpass would ever run anywhere else but
    # in Switzerland ;)
    timezone = 'Europe/Zurich'

    #: The public id of this period
    id = Column(UUID, primary_key=True, default=uuid4)

    #: The public title of this period
    title = Column(Text, nullable=False)

    #: Only one period is active at a time
    active = Column(Boolean, nullable=False, default=False)

    #: A confirmed period may not be automatically matched anymore and all
    #: booking changes to it are communicted to the customer
    confirmed = Column(Boolean, nullable=False, default=False)

    #: A finalized period may not have any change in bookings anymore
    finalized = Column(Boolean, nullable=False, default=False)

    #: An archived period has been entirely completed
    archived = Column(Boolean, nullable=False, default=False)

    #: Start of the wishlist-phase
    prebooking_start = Column(Date, nullable=False)

    #: End of the wishlist-phase
    prebooking_end = Column(Date, nullable=False)

    #: Date of the earliest possible occasion start of this period
    execution_start = Column(Date, nullable=False)

    #: Date of the latest possible occasion end of this period
    execution_end = Column(Date, nullable=False)

    #: Extra data stored on the period
    data = Column(JSON, nullable=False, default=dict)

    #: Maximum number of bookings per attendee
    max_bookings_per_attendee = Column(Integer, nullable=True)

    #: Base cost for one or many bookings
    booking_cost = Column(Numeric(precision=8, scale=2), nullable=True)

    #: True if the booking cost is meant for all bookings in a period
    #: or for each single booking
    all_inclusive = Column(Boolean, nullable=False, default=False)

    #: True if the costs of an occasions need to be paid to the organiser
    pay_organiser_directly = Column(Boolean, nullable=False, default=False)

    #: Time between bookings in minutes
    minutes_between = Column(Integer, nullable=True, default=0)

    #: The alignment of bookings in the matching
    alignment = Column(Text, nullable=True)

    #: Date after which no bookings are possible anymore
    deadline_date = Column(Date, nullable=True)

    #: Days between the occasion and the deadline (an alternative to
    #: the deadline_date)
    deadline_days = Column(Integer, nullable=True)

    #: Date after which no bookings can be canceled by a mere member
    cancellation_date = Column(Date, nullable=True)

    #: Days between the occasion and the cancellation (an alternative to
    #: the cancellation_date)
    cancellation_days = Column(Integer, nullable=True)

    __table_args__ = (
        CheckConstraint((
            '"prebooking_start" <= "prebooking_end" AND '
            '"prebooking_end" <= "execution_start" AND '
            '"execution_start" <= "execution_end"'
        ), name='period_date_order'),
        Index(
            'only_one_active_period', 'active',
            unique=True, postgresql_where=column('active') == True
        )
    )

    #: The occasions linked to this period
    occasions = relationship(
        'Occasion',
        order_by='Occasion.order',
        backref='period'
    )

    #: The bookings linked to this period
    bookings = relationship(
        'Booking',
        backref='period'
    )

    @hybrid_property
    def occasion_extra_cost(self):
        """ Returns the cost that has to be added to each occasion to
        get its actual price. Only relevant if not all-inclusive.

        """
        if self.all_inclusive:
            return 0
        else:
            return self.booking_cost or 0

    @occasion_extra_cost.expression
    def occasion_extra_cost(self):
        return case([
            (Period.all_inclusive == True, 0),
            (Period.all_inclusive == False, func.coalesce(
                Period.booking_cost, 0
            ))
        ])

    def activate(self):
        """ Activates the current period, causing all occasions and activites
        to update their status and book-keeping.

        It also makes sure no other period is active.

        """
        if self.active:
            return

        session = object_session(self)
        model = self.__class__

        active_period = session.query(model)\
            .filter(model.active == True).first()

        if active_period:
            active_period.deactivate()

        # avoid triggering the only_one_active_period index constraint
        session.flush()

        self.active = True

    def deactivate(self):
        """ Deactivates the current period, causing all occasions and activites
        to update their status and book-keeping.

        """

        if not self.active:
            return

        self.active = False

    def confirm(self):
        """ Confirms the current period. """

        self.confirmed = True

        # open bookings are marked as denied during completion
        # and the booking costs are copied over permanently (so they can't
        # change anymore)
        b = object_session(self).query(Booking)
        b = b.filter(Booking.period_id == self.id)
        b = b.options(joinedload(Booking.occasion))
        b = b.options(
            defer(Booking.group_code),
            defer(Booking.attendee_id),
            defer(Booking.priority),
            defer(Booking.username),
        )

        for booking in b:
            if booking.state == 'open':
                booking.state = 'denied'

            booking.cost = booking.provisional_booking_cost(period=self)

    def archive(self):
        """ Moves all accepted activities with an occasion in this period
        into the archived state, unless there's already another occasion
        in a period newer than the current period.

        """
        assert self.confirmed and self.finalized

        self.archived = True
        self.active = False

        session = object_session(self)

        def future_periods():
            p = session.query(Period)
            p = p.order_by(desc(Period.execution_start))
            p = p.with_entities(Period.id)

            for period in p:
                if period.id == self.id:
                    break
                yield period.id

        # get the activities which have an occasion in a future period
        f = session.query(Occasion)
        f = f.with_entities(Occasion.activity_id)
        f = f.filter(Occasion.period_id.in_(tuple(future_periods())))

        # get the activities which have an occasion in the given period but
        # no occasion in any future period
        o = session.query(Occasion)
        o = o.filter(Occasion.period_id == self.id)
        o = o.filter(not_(Occasion.activity_id.in_(f.subquery())))
        o = o.options(joinedload(Occasion.activity))

        # archive those
        for occasion in o:
            if occasion.activity.state == 'accepted':
                occasion.activity.archive()

        # also archive all activities without an occasion
        w = session.query(Occasion)
        w = w.with_entities(distinct(Occasion.activity_id))

        # XXX circular import
        from onegov.activity.models.activity import Activity

        a = session.query(Activity)
        a = a.filter(not_(Activity.id.in_(w.subquery())))
        a = a.filter(Activity.state == 'accepted')

        for activity in a:
            activity.archive()

    @property
    def booking_limit(self):
        """ Returns the max_bookings_per_attendee limit if it applies. """
        return self.all_inclusive and self.max_bookings_per_attendee

    def as_local_datetime(self, day):
        return sedate.standardize_date(
            datetime(day.year, day.month, day.day, 0, 0),
            self.timezone
        )

    @property
    def phase(self):
        local = self.as_local_datetime
        today = local(date.today())

        if not self.active or today < local(self.prebooking_start):
            return 'inactive'

        if not self.confirmed:
            return 'wishlist'

        if not self.finalized:
            return 'booking'

        if today < local(self.execution_start):
            return 'payment'

        if local(self.execution_start) <= today <= local(self.execution_end):
            return 'execution'

        if today > local(self.execution_end):
            return 'archive'

    @property
    def wishlist_phase(self):
        return self.phase == 'wishlist'

    @property
    def booking_phase(self):
        return self.phase == 'booking'

    @property
    def payment_phase(self):
        return self.phase == 'payment'

    @property
    def execution_phase(self):
        return self.phase == 'execution'

    @property
    def archive_phase(self):
        return self.phase == 'archive'

    @property
    def is_prebooking_in_future(self):
        today = self.as_local_datetime(date.today())
        start = self.as_local_datetime(self.prebooking_start)

        return today < start

    @property
    def is_currently_prebooking(self):
        if not self.wishlist_phase:
            return False

        today = self.as_local_datetime(date.today())
        start = self.as_local_datetime(self.prebooking_start)
        end = self.as_local_datetime(self.prebooking_end)

        return start <= today <= end

    @property
    def is_prebooking_in_past(self):
        today = self.as_local_datetime(date.today())
        start = self.as_local_datetime(self.prebooking_start)
        end = self.as_local_datetime(self.prebooking_end)

        if today > end:
            return True

        return start <= today and not self.wishlist_phase

    @property
    def scoring(self):
        # circular import
        from onegov.activity.matching.score import Scoring

        return Scoring.from_settings(
            settings=self.data.get('match-settings', {}),
            session=object_session(self))

    @scoring.setter
    def scoring(self, scoring):
        self.data['match-settings'] = scoring.settings



from datetime import date
from decimal import Decimal
from onegov.activity.collections import InvoiceCollection
from onegov.activity.iso20022 import extract_transactions
from onegov.activity.iso20022 import match_iso_20022_to_usernames
from onegov.activity.utils import generate_xml


def test_extract_transactions(postfinance_xml):
    transactions = extract_transactions(postfinance_xml)

    t = next(transactions)
    assert t.booking_date == date(2016, 4, 30)
    assert t.valuta_date == date(2016, 4, 30)
    assert t.amount == Decimal('24.00')
    assert t.currency == 'CHF'
    assert t.debitor is None
    assert t.debitor_account is None
    assert t.credit is False
    assert t.reference is None
    assert t.booking_text \
        == "UEBRIGE: 25-9034-2 FÜR KONTOAUSZUG PAPIER TÄGLICH"

    t = next(transactions)
    assert t.amount == Decimal('328.75')
    assert t.tid == "20160430001001080060699000901107"

    t = next(transactions)
    assert t.amount == Decimal('200.80')

    t = next(transactions)
    assert t.amount == Decimal('638.15')

    t = next(transactions)
    assert t.amount == Decimal('24.00')

    t = next(transactions)
    assert t.amount == Decimal('35.72')

    t = next(transactions)
    assert t.amount == Decimal('100.00')
    assert t.reference == "000000000002015110002913192"

    t = next(transactions)
    assert t.amount == Decimal('50.00')
    assert t.debitor == "Rutschmann Pia"
    assert t.debitor_account is None

    t = next(transactions)
    assert t.amount == Decimal('200.00')

    t = next(transactions)
    assert t.amount == Decimal('50.00')

    t = next(transactions)
    assert t.amount == Decimal('100.00')
    assert t.credit is True
    assert t.reference == "000000000002016030000535990"
    assert t.tid == "160527CH00T3L0NC"

    t = next(transactions)
    assert t.amount == Decimal('200.00')
    assert t.debitor == "Bernasconi Maria"
    assert t.debitor_account == "CH4444444444444444444"

    t = next(transactions)
    assert t.amount == Decimal('50.00')

    t = next(transactions)
    assert t.amount == Decimal('100.00')

    t = next(transactions)
    assert t.amount == Decimal('50.00')

    t = next(transactions)
    assert t.amount == Decimal('100.00')

    t = next(transactions)
    assert t.amount == Decimal('50.00')
    assert t.note == "MUSTER MITTEILUNG"

    t = next(transactions)
    assert t.amount == Decimal('20.00')

    t = next(transactions)
    assert t.amount == Decimal('100.00')
    assert t.tid == "160527CH00T3HDBA"
    assert t.note == '\n'.join([
        "MUSTER MITTEILUNG 1",
        "MUSTER MITTEILUNG 2",
        "MUSTER MITTEILUNG 3",
        "MUSTER MITTEILUNG 4",
    ])


def test_unique_transaction_ids(postfinance_xml):
    seen = set()

    for transaction in extract_transactions(postfinance_xml):
        assert transaction.tid not in seen
        seen.add(transaction.tid)


def test_invoice_matching(session, owner, member,
                          prebooking_period, inactive_period):

    p1 = prebooking_period
    p2 = inactive_period

    invoices = InvoiceCollection(session)

    own_i1 = invoices.add(user_id=owner.id, period_id=p1.id)
    mem_i1 = invoices.add(user_id=member.id, period_id=p1.id)
    mem_i2 = invoices.add(user_id=member.id, period_id=p2.id)

    i1 = own_i1.add('Aaron', 'Billard', 250, 1)
    i2 = own_i1.add('Baron', 'Pokemon', 250, 1)
    i3 = mem_i1.add('Connie', 'Swimming', 250, 1)
    i4 = mem_i2.add('Donnie', 'Football', 250, 1)

    # match against a perfect match and one with a wrong amount
    xml = generate_xml([
        dict(amount='500.00 CHF', note=i1.invoice.references[0].reference),
        dict(amount='500.00 CHF', note=i3.invoice.references[0].reference)
    ])

    transactions = list(match_iso_20022_to_usernames(xml, session, p1.id))

    assert len(transactions) == 2
    assert transactions[0].amount == Decimal(500.00)
    assert transactions[0].username == owner.username
    assert transactions[0].note \
        == i1.invoice.references[0].reference \
        == i2.invoice.references[0].reference
    assert transactions[0].confidence == 1.0
    assert transactions[1].amount == Decimal(500.00)
    assert transactions[1].username == member.username
    assert transactions[1].note == i3.invoice.references[0].reference
    assert transactions[1].confidence == 0.5

    # debit transactions are simply ignored
    xml = generate_xml([
        dict(amount='-500.00 CHF', note=i1.invoice.references[0].reference),
        dict(amount='-500.00 CHF', note=i3.invoice.references[0].reference)
    ])

    transactions = list(match_iso_20022_to_usernames(xml, session, p1.id))

    assert len(transactions) == 0

    # match against a code with extra information
    xml = generate_xml([
        dict(
            amount='500.00 CHF',
            note='Code: ' + i1.invoice.references[0].reference
        ),
    ])

    transactions = list(match_iso_20022_to_usernames(xml, session, p1.id))

    assert len(transactions) == 1
    assert transactions[0].username == owner.username
    assert transactions[0].confidence == 1.0

    # match against transactions from a different invoice
    xml = generate_xml([
        dict(amount='500.00 CHF', note=i1.invoice.references[0].reference),
        dict(amount='250.00 CHF', note=i3.invoice.references[0].reference),
        dict(amount='250.00 CHF', note=i4.invoice.references[0].reference),
    ])

    transactions = list(match_iso_20022_to_usernames(xml, session, p2.id))

    assert len(transactions) == 3
    assert transactions[0].amount == Decimal(500.00)
    assert transactions[0].username == owner.username
    assert transactions[0].confidence == 0.5
    assert transactions[1].amount == Decimal(250.00)
    assert transactions[1].username == member.username
    assert transactions[1].confidence == 0.5
    assert transactions[2].amount == Decimal(250.00)
    assert transactions[2].username == member.username
    assert transactions[2].confidence == 1.0

    # match against the wrong currency
    xml = generate_xml([
        dict(amount='250.00 EUR', note=i4.invoice.references[0].reference),
    ])

    transactions = list(match_iso_20022_to_usernames(xml, session, p2.id))

    assert len(transactions) == 1
    assert transactions[0].username is None
    assert transactions[0].confidence == 0

    # match against the wrong code
    xml = generate_xml([
        dict(amount='250.00 CHF', note='asdf'),
    ])

    transactions = list(match_iso_20022_to_usernames(xml, session, p2.id))
    assert len(transactions) == 1
    assert transactions[0].username == member.username
    assert transactions[0].confidence == 0.5

    # match against the wrong code and wrong amount
    xml = generate_xml([
        dict(amount='123.00 CHF', note='asdf'),
    ])

    transactions = list(match_iso_20022_to_usernames(xml, session, p2.id))
    assert len(transactions) == 1
    assert transactions[0].username is None
    assert transactions[0].confidence == 0

    # match against duplicate bookings
    xml = generate_xml([
        dict(amount='250.00 CHF', note=i4.invoice.references[0].reference),
        dict(amount='250.00 CHF', note=i4.invoice.references[0].reference),
    ])

    transactions = list(match_iso_20022_to_usernames(xml, session, p2.id))
    assert len(transactions) == 2
    assert transactions[0].username == member.username
    assert transactions[0].confidence == 1.0
    assert transactions[0].duplicate is True
    assert transactions[0].username == member.username
    assert transactions[0].confidence == 1.0
    assert transactions[0].duplicate is True

    # match against split bookings
    xml = generate_xml([
        dict(amount='125.00 CHF', note=i1.invoice.references[0].reference),
        dict(amount='125.00 CHF', note=i1.invoice.references[0].reference),
    ])

    transactions = list(match_iso_20022_to_usernames(xml, session, p2.id))
    assert len(transactions) == 2
    assert transactions[0].username == owner.username
    assert transactions[0].confidence == 0.5
    assert transactions[0].duplicate is True
    assert transactions[1].username == owner.username
    assert transactions[1].confidence == 0.5
    assert transactions[1].duplicate is True

    # match against reference numbers
    xml = generate_xml([
        dict(
            amount='500.00 CHF',
            reference=i1.invoice.references[0].reference
        ),
    ])

    transactions = list(match_iso_20022_to_usernames(xml, session, p1.id))

    assert len(transactions) == 1
    assert transactions[0].username == owner.username
    assert transactions[0].confidence == 1.0

    # match against a mix of reference numbers and codes
    xml = generate_xml([
        dict(
            amount='500.00 CHF',
            reference=i1.invoice.references[0].reference),
        dict(
            amount='250.00 CHF',
            note=i3.invoice.references[0].reference),
    ])

    transactions = list(match_iso_20022_to_usernames(xml, session, p1.id))

    assert len(transactions) == 2
    assert transactions[0].username == owner.username
    assert transactions[0].confidence == 1.0
    assert transactions[1].username == member.username
    assert transactions[1].confidence == 1.0

    # ignore invalid esr references
    xml = generate_xml([
        dict(amount='500.00 CHF', reference='booyah'),
        dict(amount='250.00 CHF', note=i3.invoice.references[0].reference),
    ])

    transactions = list(match_iso_20022_to_usernames(xml, session, p1.id))

    assert len(transactions) == 2
    assert transactions[0].username == owner.username
    assert transactions[0].confidence == 0.5
    assert transactions[1].username == member.username
    assert transactions[1].confidence == 1.0

    # if a single transaction has both reference and code, then one is
    # allowed to be faulty without a confidence penalty
    xml = generate_xml([
        dict(
            amount='500.00 CHF',
            reference=i1.invoice.references[0].reference,
            note=i1.invoice.references[0].reference
        ),
    ])

    transactions = list(match_iso_20022_to_usernames(xml, session, p1.id))

    assert len(transactions) == 1
    assert transactions[0].username == owner.username
    assert transactions[0].confidence == 1.0

    xml = generate_xml([
        dict(
            amount='500.00 CHF',
            reference='000127131108141601011502061',
            note=i1.invoice.references[0].reference
        ),
    ])

    transactions = list(match_iso_20022_to_usernames(xml, session, p1.id))

    assert len(transactions) == 1
    assert transactions[0].username == owner.username
    assert transactions[0].confidence == 1.0

    xml = generate_xml([
        dict(
            amount='500.00 CHF',
            reference=i1.invoice.references[0].reference,
            note='qeb3afd0e43'
        ),
    ])

    transactions = list(match_iso_20022_to_usernames(xml, session, p1.id))

    assert len(transactions) == 1
    assert transactions[0].username == owner.username
    assert transactions[0].confidence == 1.0

    # make sure paid transactions are matched as well (here we basically send
    # a separate bill which completes the outstanding amount)
    i1.paid = True
    i1.tid = 'foobar'
    i1.source = 'xml'

    xml = generate_xml([
        dict(
            amount='250 CHF',
            note=i1.invoice.references[0].reference,
            tid='foobar'),
        dict(
            amount='250 CHF',
            note=i1.invoice.references[0].reference,
            tid=None),
    ])

    transactions = list(match_iso_20022_to_usernames(xml, session, p1.id))
    assert len(transactions) == 2

    assert transactions[0].tid == 'foobar'
    assert transactions[0].username == owner.username
    assert transactions[0].paid is True

    assert transactions[1].username == owner.username
    assert transactions[1].confidence == 1.0
    assert transactions[1].paid is False


def test_invoice_matching_multischema(session, owner, prebooking_period):
    period = prebooking_period

    # create an invoice with two possible references
    invoices = InvoiceCollection(
        session,
        period_id=period.id,
        user_id=owner.id,
        schema='feriennet-v1')

    i = invoices.add()
    i.add('Aaron', 'Billard', 250, 1)

    invoices.for_schema('esr-v1').schema.link(session, i)

    # make sure we can match against both
    xml = generate_xml([
        dict(
            amount='250.00 CHF',
            note=i.references[0].reference
        ),
    ])

    transactions = list(match_iso_20022_to_usernames(xml, session, period.id))

    assert len(transactions) == 1
    assert transactions[0].amount == Decimal(250)
    assert transactions[0].username == owner.username
    assert transactions[0].confidence == 1.0

    xml = generate_xml([
        dict(
            amount='250.00 CHF',
            reference=i.references[1].reference
        ),
    ])

    transactions = list(match_iso_20022_to_usernames(xml, session, period.id))

    assert len(transactions) == 1
    assert transactions[0].amount == Decimal(250)
    assert transactions[0].username == owner.username
    assert transactions[0].confidence == 1.0



import logging
log = logging.getLogger('onegov.activity')  # noqa
log.addHandler(logging.NullHandler())  # noqa

from onegov.activity.models import (
    Activity,
    Attendee,
    Booking,
    Invoice,
    InvoiceItem,
    InvoiceReference,
    Occasion,
    OccasionDate,
    OccasionNeed,
    Period,
    PublicationRequest
)
from onegov.activity.collections import (
    ActivityCollection,
    ActivityFilter,
    AttendeeCollection,
    BookingCollection,
    InvoiceCollection,
    OccasionCollection,
    PeriodCollection,
    PublicationRequestCollection
)


__all__ = [
    'Activity',
    'ActivityFilter',
    'Attendee',
    'Booking',
    'Invoice',
    'InvoiceItem',
    'InvoiceReference',
    'Occasion',
    'OccasionDate',
    'OccasionNeed',
    'Period',
    'ActivityCollection',
    'AttendeeCollection',
    'BookingCollection',
    'InvoiceCollection',
    'OccasionCollection',
    'PeriodCollection',
    'PublicationRequest',
    'PublicationRequestCollection',
]



import sys

from datetime import date, timedelta, datetime
from functools import partial
from itertools import count
from onegov.activity.utils import dates_overlap
from onegov.activity.matching import deferred_acceptance
from onegov.activity.matching import MatchableBooking
from onegov.activity.matching import MatchableOccasion
from onegov.activity.matching import PreferAdminChildren
from onegov.activity.matching import PreferInAgeBracket
from onegov.activity.matching import PreferMotivated
from onegov.activity.matching import PreferOrganiserChildren
from onegov.activity.matching import Scoring
from onegov.activity.matching.core import is_stable, OccasionAgent
from onegov.activity.matching.utils import unblockable
from onegov.core.utils import Bunch
from sedate import standardize_date
from uuid import uuid4


today = date.today
keycount = count(start=1, step=1)

match = partial(
    deferred_acceptance,
    stability_check=True,
    validity_check=True,
    hard_budget=True
)


def days(n):
    return timedelta(days=n)


class Booking(MatchableBooking):

    def __init__(self, occasion, attendee, state, priority, dates, score=0):
        self.occasion = occasion
        self.attendee = attendee
        self._id = next(keycount)
        self._state = state
        self._priority = priority
        self._dates = dates
        self.score = score

    def __eq__(self, other):
        return self._id == other._id

    def __hash__(self):
        return hash(self._id)

    @property
    def id(self):
        return self._id

    @property
    def occasion_id(self):
        return self.occasion.id

    @property
    def attendee_id(self):
        return self.attendee

    @property
    def state(self):
        return self._state

    @property
    def priority(self):
        return self._priority

    @property
    def group_code(self):
        return uuid4().hex

    @property
    def dates(self):
        return [Bunch(start=s, end=e) for s, e in self._dates]

    def overlaps(self, other):
        if self.id == other.id:
            return True

        return dates_overlap(
            tuple((d.start, d.end) for d in self.dates),
            tuple((o.start, o.end) for o in other.dates)
        )


class Occasion(MatchableOccasion):

    def __init__(self, name, dates, max_spots=10, no_overlap_check=False,
                 anti_affinity_group=None):
        self.name = name
        self._max_spots = max_spots
        self._no_overlap_check = no_overlap_check
        self._anti_affinity_group = anti_affinity_group

        def standardize(dt):
            if isinstance(dt, datetime):
                return standardize_date(dt, 'Europe/Zurich')
            return dt

        self._dates = [(standardize(s), standardize(e)) for s, e in dates]

    @property
    def id(self):
        return self.name

    @property
    def max_spots(self):
        return self._max_spots

    def booking(self, attendee, state, priority):
        return Booking(self, attendee, state, priority, self._dates)

    @property
    def dates(self):
        return [Bunch(start=s, end=e) for s, e in self._dates]

    @property
    def exclude_from_overlap_check(self):
        return self._no_overlap_check

    @property
    def anti_affinity_group(self):
        return self._anti_affinity_group


def test_overlapping_bookings():

    # the algorithm will block other bookings, favoring higher priorities
    o1 = Occasion("Daytrip", [[today(), today()]])
    o2 = Occasion("Camp", [[today(), today() + days(1)]])
    o3 = Occasion("Zoo", [[today() + days(1), today() + days(2)]])

    bookings = [
        o1.booking("Justin", 'open', 1),
        o2.booking("Justin", 'open', 0)
    ]

    result = match(bookings, (o1, o2))

    assert not result.open
    assert result.accepted == {bookings[0]}
    assert result.blocked == {bookings[1]}

    # the order of the bookings doesn't matter
    bookings = [
        o1.booking("Justin", 'open', 0),
        o2.booking("Justin", 'open', 1)
    ]

    result = match(bookings, (o1, o2))

    assert not result.open
    assert result.blocked == {bookings[0]}
    assert result.accepted == {bookings[1]}

    # if the bookings found are instable, the algorithm will correct that
    bookings = [
        o1.booking("Justin", 'blocked', 1),
        o2.booking("Justin", 'accepted', 0)
    ]

    result = match(bookings, (o1, o2))

    assert not result.open
    assert result.accepted == {bookings[0]}
    assert result.blocked == {bookings[1]}

    # always prefer the booking with the highest priority, even if it leads
    # to more blocked bookings than it would otherwise
    bookings = [
        o1.booking("Justin", 'open', 0),
        o2.booking("Justin", 'open', 1),
        o3.booking("Justin", 'open', 0)
    ]

    result = match(bookings, (o1, o2, o3))

    assert not result.open
    assert result.blocked == {bookings[0], bookings[2]}
    assert result.accepted == {bookings[1]}

    # be predictable if there are no other preferences
    bookings = [
        o1.booking("Justin", 'open', 0),
        o2.booking("Justin", 'open', 0),
        o3.booking("Justin", 'open', 0)
    ]

    result = match(bookings, (o1, o2, o3))

    assert not result.open
    assert result.accepted == {bookings[0], bookings[2]}
    assert result.blocked == {bookings[1]}


def test_overlapping_bookings_with_multiple_dates():

    o1 = Occasion("Daytrip", [
        [today(), today()],
        [today() + days(1), today() + days(1)]
    ])

    o2 = Occasion("Camp", [
        [today() + days(1), today() + days(1)],
        [today() + days(2), today() + days(2)]
    ])

    bookings = [
        o1.booking("Justin", 'open', 1),
        o2.booking("Justin", 'open', 0)
    ]

    result = match(bookings, (o1, o2))

    assert not result.open
    assert result.accepted == {bookings[0]}
    assert result.blocked == {bookings[1]}


def test_overlap_exclusion():

    o1 = Occasion("A", [[today(), today()]])
    o2 = Occasion("B", [[today(), today()]], no_overlap_check=True)

    bookings = [
        o1.booking("Justin", 'open', 1),
        o2.booking("Justin", 'open', 0)
    ]

    result = match(bookings, (o1, o2))
    assert result.accepted == {bookings[0], bookings[1]}


def test_overlapping_bookings_with_minutes_between():

    o1 = Occasion("A", [[
        datetime(2017, 2, 16, 10),
        datetime(2017, 2, 16, 11),
    ]])

    o2 = Occasion("B", [[
        datetime(2017, 2, 16, 11),
        datetime(2017, 2, 16, 12),
    ]])

    o3 = Occasion("C", [[
        datetime(2017, 2, 16, 12),
        datetime(2017, 2, 16, 13),
    ]])

    bookings = [
        o1.booking("Justin", 'open', 2),
        o2.booking("Justin", 'open', 1),
        o3.booking("Justin", 'open', 0),
    ]

    result = match(bookings, (o1, o2, o3))
    assert not result.open
    assert result.accepted == {bookings[0], bookings[1], bookings[2]}

    result = match(bookings, (o1, o2, o3), minutes_between=1)
    assert not result.open
    assert result.accepted == {bookings[0], bookings[2]}
    assert result.blocked == {bookings[1]}

    result = match(bookings, (o1, o2, o3), minutes_between=60)
    assert not result.open
    assert result.accepted == {bookings[0], bookings[2]}
    assert result.blocked == {bookings[1]}

    result = match(bookings, (o1, o2, o3), minutes_between=61)
    assert not result.open
    assert result.accepted == {bookings[0]}
    assert result.blocked == {bookings[1], bookings[2]}

    bookings = [
        o1.booking("Justin", 'open', 1),
        o2.booking("Justin", 'open', 2),
        o3.booking("Justin", 'open', 0),
    ]

    result = match(bookings, (o1, o2, o3), minutes_between=1)
    assert not result.open
    assert result.accepted == {bookings[1]}
    assert result.blocked == {bookings[0], bookings[2]}


def test_is_stable():

    o1 = Occasion("A", [[
        datetime(2017, 2, 16, 10),
        datetime(2017, 2, 16, 11),
    ]])

    o2 = Occasion("B", [[
        datetime(2017, 2, 16, 11),
        datetime(2017, 2, 16, 12),
    ]])

    o3 = Occasion("C", [[
        datetime(2017, 2, 16, 13),
        datetime(2017, 2, 16, 14),
    ]])

    bookings = [
        o1.booking("Justin", 'open', 1),
        o2.booking("Justin", 'open', 0)
    ]

    result = match(bookings, (o1, o2))

    attendees = [Bunch(accepted=result.accepted)]

    def preferred(b):
        call_frame = sys._getframe(1)
        occasion = call_frame.f_locals['self']
        return b.preferred_occasion == occasion.id and 1 or 0

    # if no bookings prefer another occasion which prefers them -> stable
    occasions = [OccasionAgent(o, preferred) for o in (o1, o2)]
    occasions[0].bookings.add(bookings[0])
    occasions[1].bookings.add(bookings[1])

    bookings[0].preferred_occasion = 'A'
    bookings[1].preferred_occasion = 'B'

    assert is_stable(attendees, occasions)

    # if there are bookings which would like to be swapped -> unstable
    occasions = [OccasionAgent(o, preferred) for o in (o1, o2)]

    occasions[0].bookings.add(bookings[0])
    occasions[1].bookings.add(bookings[1])

    bookings[0].preferred_occasion = 'B'
    bookings[1].preferred_occasion = 'A'

    assert not is_stable(attendees, occasions)

    # if there are multiple bookings where some would like to be swapped,
    # but not vice versa (here the priority creates a cascade) -> stable
    bookings = [
        o1.booking("Justin", 'open', 2),
        o2.booking("Justin", 'open', 1),
        o3.booking("Justin", 'open', 0)
    ]

    result = match(bookings, (o1, o2, o3))
    attendees = [Bunch(accepted=result.accepted)]

    occasions = [OccasionAgent(o, lambda b: b.priority) for o in (o1, o2, o3)]
    occasions[0].bookings.add(bookings[0])
    occasions[1].bookings.add(bookings[1])
    occasions[2].bookings.add(bookings[2])

    assert is_stable(attendees, occasions)

    # multiple bookings which like to be swapped (but not all) -> unstable

    occasions = [OccasionAgent(o, preferred) for o in (o1, o2, o3)]
    occasions[0].bookings.add(bookings[0])
    occasions[1].bookings.add(bookings[1])
    occasions[2].bookings.add(bookings[2])

    bookings[0].preferred_occasion = 'A'
    bookings[1].preferred_occasion = 'C'
    bookings[2].preferred_occasion = 'B'

    assert not is_stable(attendees, occasions)


def test_accept_highest_priority():

    o = Occasion("Best Activity Ever", [[today(), today()]], max_spots=2)
    bookings = [
        o.booking("Tick", 'open', 0),
        o.booking("Trick", 'open', 1),
        o.booking("Track", 'open', 1)
    ]

    result = match(bookings, [o])

    assert result.open == {bookings[0]}
    assert result.accepted == {bookings[1], bookings[2]}
    assert not result.blocked

    # all other things being equal, the choice is inherently random.
    # to be predictable we need to change all the set and dictionaries to
    # their ordered counterparts which is something we want to avoid for
    # performance/memory usage reasons
    o = Occasion("Best Activity Ever", [[today(), today()]], max_spots=2)
    bookings = [
        o.booking("Tick", 'open', 0),
        o.booking("Trick", 'open', 0),
        o.booking("Track", 'open', 1)
    ]

    result = match(bookings, [o])

    # the first result loses, because it's the first result with a lower
    # score than the third booking - we could turn this around, the important
    # thing is that we are predictable
    assert result.open == {bookings[0]} or result.open == {bookings[1]}
    assert result.accepted == {bookings[1], bookings[2]} or\
        result.accepted == {bookings[0], bookings[2]}
    assert not result.blocked


def test_prefer_motivated():
    motivation_score = PreferMotivated()

    assert motivation_score(Bunch(priority=1)) == 1
    assert motivation_score(Bunch(priority=0)) == 0
    assert motivation_score(Bunch(priority=123)) == 123


def test_prefer_in_age_bracket():
    age_range = None
    attendee_age = None

    age_bracket_score = PreferInAgeBracket(
        get_age_range=lambda b: age_range,
        get_attendee_age=lambda b: attendee_age)

    age_range = (10, 20)
    attendee_age = 10

    assert age_bracket_score(None) == 1.0

    attendee_age = 21
    assert age_bracket_score(None) == 0.9

    attendee_age = 22
    assert age_bracket_score(None) == 0.8

    attendee_age = 23
    assert age_bracket_score(None) == 0.7

    attendee_age = 30
    assert age_bracket_score(None) == 0.0

    attendee_age = 9
    assert age_bracket_score(None) == 0.9

    attendee_age = 8
    assert age_bracket_score(None) == 0.8


def test_prefer_organiser_children():

    is_organiser_child = None

    organiser_child_score = PreferOrganiserChildren(
        get_is_organiser_child=lambda c: is_organiser_child)

    is_organiser_child = True
    assert organiser_child_score(None) == 1.0

    is_organiser_child = False
    assert organiser_child_score(None) == 0.0


def test_prefer_association_children():

    is_association_child = None

    association_child_score = PreferAdminChildren(
        get_is_association_child=lambda c: is_association_child)

    is_association_child = True
    assert association_child_score(None) == 1.0

    is_association_child = False
    assert association_child_score(None) == 0.0


def test_serialize_scoring(session):

    scoring = Scoring()
    assert scoring.settings == {}

    scoring.criteria.append(PreferInAgeBracket.from_session(session))
    assert scoring.settings == {
        'prefer_in_age_bracket': True
    }

    scoring.criteria.append(PreferOrganiserChildren.from_session(session))
    assert scoring.settings == {
        'prefer_in_age_bracket': True,
        'prefer_organiser': True
    }

    scoring.criteria.append(PreferAdminChildren.from_session(session))
    assert scoring.settings == {
        'prefer_in_age_bracket': True,
        'prefer_organiser': True,
        'prefer_admins': True
    }

    scoring = Scoring.from_settings(scoring.settings, session)

    assert len(scoring.criteria) == 5


def test_booking_limit():

    o1 = Occasion(1, [[today(), today()]])
    o2 = Occasion(2, [[today() + days(1), today() + days(1)]])
    o3 = Occasion(3, [[today() + days(2), today() + days(2)]])
    o4 = Occasion(4, [[today() + days(3), today() + days(3)]])

    bookings = [
        o1.booking("Tom", 'open', 0),
        o2.booking("Tom", 'open', 0),
        o3.booking("Tom", 'open', 0),
        o4.booking("Tom", 'open', 0)
    ]

    result = match(bookings, (o1, o2, o3, o4), default_limit=1)

    assert not result.open
    assert result.accepted == {bookings[0]}
    assert result.blocked == {bookings[1], bookings[2], bookings[3]}

    result = match(bookings, (o1, o2, o3, o4), default_limit=2)

    assert not result.open
    assert result.accepted == {bookings[0], bookings[1]}
    assert result.blocked == {bookings[2], bookings[3]}

    result = match(bookings, (o1, o2, o3, o4), default_limit=3)

    assert not result.open
    assert result.accepted == {bookings[0], bookings[1], bookings[2]}
    assert result.blocked == {bookings[3]}

    result = match(bookings, (o1, o2, o3, o4), default_limit=4)

    assert not result.open
    assert len(result.accepted) == 4
    assert not result.blocked

    bookings = [
        o1.booking("Tom", 'open', 0),
        o2.booking("Tom", 'open', 0),
        o3.booking("Tom", 'open', 1),
        o4.booking("Tom", 'open', 0)
    ]

    result = match(bookings, (o1, o2, o3, o4), default_limit=1)

    assert not result.open
    assert result.accepted == {bookings[2]}
    assert result.blocked == {bookings[0], bookings[1], bookings[3]}

    bookings = [
        o1.booking("Tom", 'open', 2),
        o2.booking("Tom", 'open', 1),
        o3.booking("Tom", 'open', 0),
        o1.booking("Harry", 'open', 0),
        o2.booking("Harry", 'open', 2),
        o3.booking("Harry", 'open', 1),
    ]

    result = match(bookings, (o1, o2, o3, o4), default_limit=1)

    assert not result.open
    assert result.accepted == {bookings[0], bookings[4]}
    assert len(result.blocked) == 4

    result = match(bookings, (o1, o2, o3, o4), default_limit=2)

    assert not result.open
    assert result.accepted == {
        bookings[0],
        bookings[1],
        bookings[4],
        bookings[5]
    }
    assert len(result.blocked) == 2

    bookings = [
        o1.booking("Tom", 'open', 1),
        o2.booking("Tom", 'open', 0),
        o1.booking("Dick", 'open', 1),
        o2.booking("Dick", 'open', 0),
        o1.booking("Harry", 'open', 1),
        o2.booking("Harry", 'open', 0),
    ]

    result = match(bookings, (o1, o2), default_limit=2, attendee_limits={
        "Tom": 1,
        "Dick": 1,
    })

    assert len(result.accepted) == 4
    assert result.accepted == {
        bookings[0],
        bookings[2],
        bookings[4],
        bookings[5],
    }


def test_day_alignment():
    o1 = Occasion(1, [
        [datetime(2017, 2, 20, 8), datetime(2017, 2, 20, 16)]
    ])
    o2 = Occasion(2, [
        [datetime(2017, 2, 21, 8), datetime(2017, 2, 21, 16)],
        [datetime(2017, 2, 22, 8), datetime(2017, 2, 22, 12)]
    ])
    o3 = Occasion(3, [
        [datetime(2017, 2, 22, 13), datetime(2017, 2, 22, 16)]
    ])

    bookings = [
        o1.booking("Tom", 'open', 0),
        o2.booking("Tom", 'open', 0),
        o3.booking("Tom", 'open', 0)
    ]

    # o2 and o3 do not overlap, but they have an occasion on the same day
    # which should be catpured by our occasions_per_day limit
    result = match(bookings, (o1, o2, o3))
    assert len(result.accepted) == 3

    result = match(bookings, (o1, o2, o3), alignment='day')
    assert len(result.accepted) == 2


def test_split_day_alignment():
    o1 = Occasion(1, [
        [datetime(2017, 2, 20, 8), datetime(2017, 2, 20, 16)]
    ])
    o2 = Occasion(2, [
        [datetime(2017, 2, 21, 8), datetime(2017, 2, 21, 12)],
        [datetime(2017, 2, 21, 13), datetime(2017, 2, 21, 16)]
    ])

    bookings = [
        o1.booking("Tom", 'open', 0),
        o2.booking("Tom", 'open', 0),
    ]

    # the occasion o2 is split into two occasions on the same day ->
    # we however do not consider this to be a block, since it's the same
    # occasion
    assert len(match(bookings, (o1, o2)).accepted) == 2
    assert len(match(bookings, (o1, o2), alignment='day').accepted) == 2


def test_multi_day_alignment():
    o1 = Occasion(1, [
        [datetime(2017, 2, 20, 8), datetime(2017, 2, 21, 16)]
    ])
    o2 = Occasion(2, [
        [datetime(2017, 2, 21, 17), datetime(2017, 2, 21, 18)]
    ])

    bookings = [
        o1.booking("Tom", 'open', 0),
        o2.booking("Tom", 'open', 0),
    ]

    # the occasion o2 is split into two occasions on the same day ->
    # we however do not consider this to be a block, since it's the same
    # occasion
    assert len(match(bookings, (o1, o2)).accepted) == 2
    assert len(match(bookings, (o1, o2), alignment='day').accepted) == 1


def test_unblockable_regression():
    def booking(start, end):
        start = datetime(2018, 3, 8, *start)
        end = datetime(2018, 3, 8, *end)

        return Booking(
            occasion=None,
            attendee=None,
            state=None,
            priority=0,
            dates=[(start, end)]
        )

    accepted = {
        booking((7, 0), (8, 0)),
        booking((8, 0), (13, 0))
    }

    blocked = {
        booking((7, 0), (9, 0))
    }

    assert not unblockable(accepted, blocked)


def test_anti_affinity_groups():
    foo = Occasion("foo", [
        (datetime(2019, 2, 12, 8), datetime(2019, 2, 12, 16))
    ])
    bar = Occasion("bar", [
        (datetime(2019, 2, 13, 8), datetime(2019, 2, 13, 16))
    ])

    bookings = [
        foo.booking("Tom", 'open', 0),
        bar.booking("Tom", 'open', 1)
    ]

    foo._anti_affinity_group = None
    bar._anti_affinity_group = None
    assert len(match(bookings, (foo, bar)).accepted) == 2

    foo._anti_affinity_group = 'xyz'
    bar._anti_affinity_group = 'zyx'
    assert len(match(bookings, (foo, bar)).accepted) == 2

    foo._anti_affinity_group = 'xyz'
    bar._anti_affinity_group = 'xyz'
    assert len(match(bookings, (foo, bar)).accepted) == 1

    # the overlap check is less important than the anti-affinity
    foo._no_overlap_check = True
    bar._no_overlap_check = True
    assert len(match(bookings, (foo, bar)).accepted) == 1



from onegov.core.orm import Base
from onegov.core.orm.mixins import TimestampMixin
from onegov.core.orm.types import UUID
from onegov.pay import PayableManyTimes
from sqlalchemy import Boolean
from sqlalchemy import Column
from sqlalchemy import ForeignKey
from sqlalchemy import Numeric
from sqlalchemy import Text
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import validates
from uuid import uuid4


# total digits
PRECISION = 8

# digits after the point
SCALE = 2


class InvoiceItem(Base, TimestampMixin, PayableManyTimes):
    """ An item in an invoice. """

    __tablename__ = 'invoice_items'

    #: the public id of the invoice item
    id = Column(UUID, primary_key=True, default=uuid4)

    #: the invoice this item belongs to
    invoice_id = Column(UUID, ForeignKey('invoices.id'))

    #: the item group (all items with the same text are visually grouped)
    group = Column(Text, nullable=False)

    #: a secondary group who is not necessarily grouped visually
    family = Column(Text, nullable=True)

    #: the item text
    text = Column(Text, nullable=False)

    #: true if paid
    paid = Column(Boolean, nullable=False, default=False)

    #: the transaction id if paid through a bank or online transaction
    tid = Column(Text, nullable=True)

    #: the source of the transaction id, e.g. stripe, xml
    source = Column(Text, nullable=True)

    #: the unit to pay..
    unit = Column(Numeric(precision=PRECISION, scale=SCALE), nullable=True)

    #: ..multiplied by the quantity..
    quantity = Column(Numeric(precision=PRECISION, scale=SCALE), nullable=True)

    #: ..together form the amount
    @hybrid_property
    def amount(self):
        return round(self.unit, SCALE) * round(self.quantity, SCALE)

    @amount.expression
    def amount(cls):
        return cls.unit * cls.quantity

    @validates('source')
    def validate_source(self, key, value):
        assert value in (None, 'xml', 'stripe_connect')
        return value



""" Contains upgrade tasks that are executed when the application is being
upgraded on the server. See :class:`onegov.core.upgrade.upgrade_task`.

"""
import hashlib
import string

from itertools import chain
from onegov.activity import ActivityCollection
from onegov.activity import Booking
from onegov.activity import InvoiceCollection
from onegov.activity import InvoiceItem
from onegov.activity import InvoiceReference
from onegov.activity import Occasion
from onegov.activity import Period
from onegov.activity import PeriodCollection
from onegov.core.crypto import random_token
from onegov.core.orm.sql import as_selectable
from onegov.core.orm.types import UUID, JSON
from onegov.core.upgrade import upgrade_task
from onegov.core.utils import Bunch
from sqlalchemy import Boolean
from sqlalchemy import Column
from sqlalchemy import Date
from sqlalchemy import desc
from sqlalchemy import Enum
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import Numeric
from sqlalchemy import select
from sqlalchemy import Text
from sqlalchemy.dialects.postgresql import ARRAY
from sqlalchemy.orm import joinedload


@upgrade_task('Rebuild bookings')
def rebuild_bookings(context):
    # having not created any bookings yet, we can rebuild them
    context.operations.drop_table('bookings')


@upgrade_task('Add period_id to bookings')
def add_period_id_to_bookings(context):

    context.operations.add_column('bookings', Column(
        'period_id', UUID, ForeignKey("periods.id"), nullable=True
    ))

    bookings = context.session.query(Booking)
    bookings = bookings.options(joinedload(Booking.occasion))

    for booking in bookings:
        booking.period_id = booking.occasion.period_id

    context.session.flush()
    context.operations.alter_column('bookings', 'period_id', nullable=False)


@upgrade_task('Change booking states')
def change_booking_states(context):

    new_type = Enum(
        'open',
        'blocked',
        'accepted',
        'denied',
        'cancelled',
        name='booking_state'
    )

    op = context.operations

    op.execute("""
        ALTER TABLE bookings ALTER COLUMN state TYPE Text;
        UPDATE bookings SET state = 'open' WHERE state = 'unconfirmed';
        DROP TYPE booking_state;
    """)

    new_type.create(op.get_bind())

    op.execute("""
        ALTER TABLE bookings ALTER COLUMN state
        TYPE booking_state USING state::text::booking_state;
    """)


@upgrade_task('Add confirmed flag to period')
def add_confirmed_flag_to_period(context):
    context.operations.add_column('periods', Column(
        'confirmed', Boolean, nullable=True, default=False
    ))

    for period in context.session.query(Period):
        period.confirmed = False

    context.session.flush()
    context.operations.alter_column('periods', 'confirmed', nullable=False)


@upgrade_task('Add data column to period')
def add_data_column_to_period(context):
    context.operations.add_column('periods', Column(
        'data', JSON, nullable=True, default=dict
    ))

    for period in context.session.query(Period):
        period.data = {}

    context.session.flush()
    context.operations.alter_column('periods', 'data', nullable=False)


@upgrade_task('Add attendee_count column to occasion')
def add_attendee_count_column_to_occasion(context):
    context.operations.add_column('occasions', Column(
        'attendee_count', Integer, nullable=True, default=0
    ))

    for occasion in context.session.query(Occasion):
        occasion.attendee_count = len(occasion.accepted)

    context.session.flush()
    context.operations.alter_column(
        'occasions', 'attendee_count', nullable=False)


@upgrade_task('Add finalized flag to period')
def add_finalized_flag_to_period(context):
    context.operations.add_column('periods', Column(
        'finalized', Boolean, nullable=True, default=False
    ))

    for period in context.session.query(Period):
        period.finalized = False

    context.session.flush()
    context.operations.alter_column('periods', 'finalized', nullable=False)


@upgrade_task('Add payment model columns')
def add_payment_model_columns(context):
    context.operations.add_column('periods', Column(
        'max_bookings_per_attendee', Integer, nullable=True
    ))

    context.operations.add_column('periods', Column(
        'booking_cost', Numeric(precision=8, scale=2), nullable=True
    ))

    context.operations.add_column('periods', Column(
        'all_inclusive', Boolean, nullable=True
    ))

    for period in context.session.query(Period):
        period.all_inclusive = False

    context.session.flush()
    context.operations.alter_column('periods', 'all_inclusive', nullable=False)

    context.operations.add_column('occasions', Column(
        'cost', Numeric(precision=8, scale=2), nullable=True
    ))

    context.operations.add_column('bookings', Column(
        'cost', Numeric(precision=8, scale=2), nullable=True
    ))


@upgrade_task('Add cancelled flag to occasion')
def add_cancelled_flag_to_occasion(context):
    context.operations.add_column('occasions', Column(
        'cancelled', Boolean, nullable=True
    ))

    for occasion in context.session.query(Occasion):
        occasion.cancelled = False

    context.session.flush()
    context.operations.alter_column('occasions', 'cancelled', nullable=False)


@upgrade_task('Add code field to invoice items')
def add_code_field_to_invoice_items(context):
    context.operations.add_column('invoice_items', Column(
        'code', Text, nullable=True
    ))

    for i in context.session.query(InvoiceItem):
        i.code = 'q' + ''.join((
            hashlib.sha1((
                i.invoice + i.username).encode('utf-8')
            ).hexdigest()[:5],
            hashlib.sha1(
                i.username.encode('utf-8')
            ).hexdigest()[:5]
        ))

    context.session.flush()
    context.operations.alter_column('invoice_items', 'code', nullable=False)


@upgrade_task('Add source field to invoice items')
def add_source_field_to_invoice_items(context):
    if not context.has_column('invoice_items', 'source'):
        context.operations.add_column('invoice_items', Column(
            'source', Text, nullable=True
        ))


@upgrade_task('Add gender/notes fields to attendees')
def add_gender_notes_fields_to_attendees(context):
    if not context.has_column('attendees', 'gender'):
        context.operations.add_column('attendees', Column(
            'gender', Text, nullable=True
        ))

    if not context.has_column('attendees', 'notes'):
        context.operations.add_column('attendees', Column(
            'notes', Text, nullable=True
        ))


@upgrade_task('Support multiple dates per occasion')
def support_multiple_dates_per_occasion(context):
    context.operations.drop_constraint('start_before_end', 'occasions')

    for name in ('durations', 'order'):
        context.operations.add_column(
            'occasions', Column(name, Integer, server_default="0"))
        context.operations.alter_column(
            'occasions', name, server_default=None)

    context.session.execute("""
        INSERT INTO occasion_dates ("timezone", "start", "end", "occasion_id")
        SELECT "timezone", "start", "end", "id" FROM occasions
    """)

    context.session.flush()

    # update dates
    for occasion in context.session.query(Occasion):
        occasion.on_date_change()

    for name in ('start', 'end', 'timezone'):
        context.operations.drop_column('occasions', name)


@upgrade_task('Adds minutes_between to period')
def adds_minutes_between_to_period(context):
    context.operations.add_column('periods', Column(
        'minutes_between', Integer, nullable=True, server_default='0'))
    context.operations.alter_column(
        'periods', 'minutes_between', server_default=None)


@upgrade_task('Adds exclude_from_overlap_check to period')
def adds_exclude_from_overlap_check_to_period(context):
    context.operations.add_column('occasions', Column(
        'exclude_from_overlap_check', Boolean, nullable=False,
        server_default='FALSE'))
    context.operations.alter_column(
        'occasions', 'exclude_from_overlap_check', server_default=None)


@upgrade_task('Adds deadlines to period')
def adds_deadlines_to_period(context):
    if not context.has_column('periods', 'deadline_date'):
        context.operations.add_column('periods', Column(
            'deadline_date', Date, nullable=True
        ))
    if not context.has_column('periods', 'deadline_days'):
        context.operations.add_column('periods', Column(
            'deadline_days', Integer, nullable=True
        ))


@upgrade_task('Adds limit to attendee')
def adds_limit_to_attendee(context):
    if not context.has_column('attendees', 'limit'):
        context.operations.add_column('attendees', Column(
            'limit', Integer, nullable=True
        ))


@upgrade_task('Introduce location/meeting_point')
def introduce_location_meeting_point(context):
    if not context.has_column('activities', 'location'):
        context.operations.add_column('activities', Column(
            'location', Text, nullable=True
        ))
    if not context.has_column('occasions', 'meeting_point'):
        context.operations.alter_column(
            'occasions', 'location', new_column_name='meeting_point')


@upgrade_task('Add active days')
def add_active_days(context):
    if context.has_column('activities', 'active_days'):
        assert context.has_column('occasions', 'active_days')
        return

    context.session.execute(
        'CREATE AGGREGATE "{}".array_cat_agg(anyarray) '
        '(SFUNC=array_cat, STYPE=anyarray)'.format(
            context.schema
        )
    )

    context.operations.add_column('activities', Column(
        'active_days', ARRAY(Integer), nullable=True
    ))

    context.operations.add_column('occasions', Column(
        'active_days', ARRAY(Integer), nullable=True
    ))

    context.session.flush()

    # update dates
    for occasion in context.session.query(Occasion):
        occasion.on_date_change()


@upgrade_task('Add active days index')
def add_active_days_index(context):
    context.operations.create_index(
        'inverted_active_days', 'activities', ['active_days'],
        postgresql_using='gin'
    )


@upgrade_task('Removed denied activity state')
def remove_denied_activity_state(context):

    new_type = Enum(
        'preview',
        'proposed',
        'accepted',
        'archived',
        name='activity_state'
    )

    op = context.operations

    op.execute("""
        ALTER TABLE activities ALTER COLUMN state TYPE Text;
        UPDATE activities SET state = 'archived' WHERE state = 'denied';
        DROP TYPE activity_state;
    """)

    new_type.create(op.get_bind())

    op.execute("""
        ALTER TABLE activities ALTER COLUMN state
        TYPE activity_state USING state::text::activity_state;
    """)


@upgrade_task('Add weekdays')
def add_weekdays(context):
    if context.has_column('activities', 'weekdays'):
        assert context.has_column('occasions', 'weekdays')
        return

    context.operations.add_column('activities', Column(
        'weekdays', ARRAY(Integer), nullable=True
    ))

    context.operations.add_column('occasions', Column(
        'weekdays', ARRAY(Integer), nullable=True
    ))

    context.session.flush()

    # update dates
    for occasion in context.session.query(Occasion):
        occasion.on_date_change()


@upgrade_task('Add weekdays index')
def add_weekdays_index(context):
    context.operations.create_index(
        'inverted_weekdays', 'activities', ['weekdays'],
        postgresql_using='gin'
    )


@upgrade_task('Extract thumbnails')
def extract_thumbnails(context):
    for activity in ActivityCollection(context.session).query():
        activity.content_observer(None)


@upgrade_task('Retroactively create publication requests')
def retroactively_create_publication_requests(context):
    activities = ActivityCollection(context.session).query()

    if not activities.count():
        return False

    periods = PeriodCollection(context.session)

    p = periods.query()
    p = p.order_by(desc(Period.active), desc(Period.execution_start))
    p = p.first()

    assert p, "an active period is required"

    for activity in activities:
        activity.create_publication_request(p, id=activity.id)


@upgrade_task('Add archived flag to period')
def add_archived_flag_to_period(context):
    context.operations.add_column('periods', Column(
        'archived', Boolean, nullable=True, default=False
    ))

    for period in context.session.query(Period):
        period.archived = False

    context.session.flush()
    context.operations.alter_column('periods', 'archived', nullable=False)


@upgrade_task('Extract municipality from activities')
def add_municipality_column_to_activites(context):
    context.operations.add_column('activities', Column(
        'municipality', Text, nullable=True
    ))

    for activity in ActivityCollection(context.session).query():
        activity.location_observer(None)


@upgrade_task('Add family to invoice items')
def add_family_to_invoice_items(context):
    context.operations.add_column('invoice_items', Column(
        'family', Text, nullable=True
    ))


@upgrade_task('Add subscription_token to attendeeds')
def add_subscription(context):
    context.add_column_with_defaults(
        table='attendees',
        column=Column('subscription_token', Text, nullable=False, unique=True),
        default=lambda attendee: random_token()
    )


@upgrade_task('Add alignment to periods')
def add_alignment(context):
    if not context.has_column('periods', 'alignment'):
        context.operations.add_column('periods', Column(
            'alignment', Text, nullable=True
        ))


@upgrade_task('Update duration categories')
def update_duration_categories(context):
    for occasion in context.session.query(Occasion):
        occasion.on_date_change()


@upgrade_task('Rename occasion durations to the singular')
def rename_occasion_durations_to_the_singular(context):
    context.operations.alter_column(
        table_name='occasions',
        column_name='durations',
        new_column_name='duration'
    )

    for occasion in context.session.query(Occasion):
        occasion.on_date_change()


@upgrade_task('Add pay_organiser_directly column')
def add_pay_organiser_directly_column(context):
    context.add_column_with_defaults(
        table='periods',
        column=Column('pay_organiser_directly', Boolean, nullable=False),
        default=False
    )


@upgrade_task('Remove activity aggregates')
def remove_activity_aggregates(context):
    context.operations.drop_index('inverted_weekdays', 'activities')
    context.operations.drop_index('inverted_active_days', 'activities')
    context.operations.drop_column('activities', 'durations')
    context.operations.drop_column('activities', 'ages')
    context.operations.drop_column('activities', 'period_ids')
    context.operations.drop_column('activities', 'active_days')
    context.operations.drop_column('activities', 'weekdays')


@upgrade_task('Add invoices')
def add_invoices(context):
    if not context.has_column('invoice_items', 'code'):
        return

    if not context.has_column('invoices', 'code'):
        context.operations.add_column('invoices', Column(
            'code', Text, nullable=True
        ))

    stmt = as_selectable("""
        SELECT
            invoice_items.id,           -- UUID
            code,                       -- Text
            invoice::uuid AS period_id, -- UUID
            users.id AS user_id         -- UUID
        FROM invoice_items
        LEFT JOIN users ON invoice_items.username = users.username
    """)

    invoices = InvoiceCollection(context.session)

    mapping = {
        r.id: Bunch(record=r, invoice=None)
        for r in context.session.execute(select(stmt.c))
    }

    created = {}

    for m in mapping.values():
        id = (m.record.period_id, m.record.user_id)

        if id in created:
            m.invoice = created[id]
            continue

        created[id] = m.invoice = invoices.add(
            period_id=m.record.period_id,
            user_id=m.record.user_id,
            code=m.record.code)

    def invoice_id(item):
        return mapping[item.id].invoice.id

    context.add_column_with_defaults(
        table='invoice_items',
        column=Column('invoice_id', UUID, ForeignKey('invoices.id')),
        default=invoice_id
    )

    context.session.flush()

    context.operations.drop_column('invoice_items', 'username')
    context.operations.drop_column('invoice_items', 'code')
    context.operations.drop_column('invoice_items', 'invoice')


@upgrade_task('Add invoice references')
def add_invoice_references(context):
    if not context.has_column('invoices', 'code'):
        return

    # legacy functions obsolete after this migration
    CODE_TO_ESR_MAPPING = {
        character: '{:02d}'.format(value) for value, character in chain(
            enumerate(string.digits, start=1),
            enumerate(string.ascii_lowercase, start=11)
        )
    }

    def generate_checksum(number):
        table = (0, 9, 4, 6, 8, 2, 7, 1, 3, 5)
        carry = 0

        for n in str(number):
            carry = table[(carry + int(n)) % 10]

        return (10 - carry) % 10

    def append_checksum(number):
        number = str(number)
        return number + str(generate_checksum(number))

    def encode_invoice_code(code):
        version = '1'

        blocks = [version]

        for char in code:
            if char in CODE_TO_ESR_MAPPING:
                blocks.append(CODE_TO_ESR_MAPPING[char])
            else:
                raise RuntimeError(
                    "Invalid character {} in {}".format(char, code))

        return append_checksum('{:0>26}'.format(''.join(blocks)))

    stmt = as_selectable("""
        SELECT
            invoices.id,           -- UUID
            invoices.code          -- Text
        FROM invoices
    """)

    invoices = context.session.execute(select(stmt.c))

    # the references might have been created already in the previous upgrade
    # step, if it was executed with the latest release
    context.session.execute('DELETE FROM invoice_references')

    for invoice in invoices:
        context.session.add(
            InvoiceReference(
                invoice_id=invoice.id,
                reference=invoice.code,
                schema='feriennet-v1',
                bucket='feriennet-v1'
            )
        )

        context.session.add(
            InvoiceReference(
                invoice_id=invoice.id,
                reference=encode_invoice_code(invoice.code),
                schema='esr-v1',
                bucket='esr-v1'
            )
        )

    context.operations.drop_column('invoices', 'code')


@upgrade_task('Add invoice reference bucket')
def add_invoice_reference_bucket(context):
    if context.has_column('invoice_references', 'bucket'):
        return

    context.add_column_with_defaults(
        table='invoice_references',
        column=Column('bucket', Text),
        default=lambda r: r.schema
    )

    context.operations.drop_constraint(
        'unique_schema_invoice_id', 'invoice_references')

    context.operations.create_unique_constraint(
        'unique_bucket_invoice_id', 'invoice_references', (
            'bucket', 'invoice_id'
        ))


@upgrade_task('Adds cancellation deadlines to period')
def adds_cancellation_deadlines_to_period(context):
    if not context.has_column('periods', 'cancellation_date'):
        context.operations.add_column('periods', Column(
            'cancellation_date', Date, nullable=True
        ))
    if not context.has_column('periods', 'cancellation_days'):
        context.operations.add_column('periods', Column(
            'cancellation_days', Integer, nullable=True
        ))


@upgrade_task('Make group_code nullable')
def make_group_code_nullable(context):
    context.operations.alter_column('bookings', 'group_code', nullable=True)

    # nobody uses groups yet, so we can safely reset it all to NULL
    context.operations.execute("UPDATE bookings SET group_code = NULL")


@upgrade_task('Adds exempt_from_booking_limit to occasion')
def adds_exempt_from_booking_limit_to_occasion(context):
    context.add_column_with_defaults(
        table='occasions',
        column=Column('exempt_from_booking_limit', Boolean, nullable=False),
        default=False
    )


@upgrade_task('Adds score to bookings')
def adds_score_to_bookings(context):
    context.add_column_with_defaults(
        table='bookings',
        column=Column('score', Numeric(precision=14, scale=9), nullable=False),
        default=0
    )


@upgrade_task('Drop occasion.active column')
def drop_occasion_active_column(context):
    context.operations.drop_column('occasions', 'active')



from onegov.core.orm import Base
from onegov.core.orm.mixins import TimestampMixin
from onegov.core.orm.types import UUID
from sqlalchemy import Column
from sqlalchemy import ForeignKey
from sqlalchemy import Text
from sqlalchemy import UniqueConstraint
from sqlalchemy.dialects.postgresql import INT4RANGE
from uuid import uuid4


class OccasionNeed(Base, TimestampMixin):
    """ Defines a required resource on an occasion. """

    __tablename__ = 'occasion_needs'

    #: the public id of this occasion resource
    id = Column(UUID, primary_key=True, default=uuid4)

    #: the name of the occasion resource
    name = Column(Text, nullable=False)

    #: a description of the occasion resource
    description = Column(Text, nullable=True)

    #: the required range of resources
    number = Column(INT4RANGE, nullable=False)

    #: The associated occasion
    occasion_id = Column(UUID, ForeignKey('occasions.id'), nullable=False)

    __table_args__ = (
        UniqueConstraint(
            'name', 'occasion_id', name='unique_need_per_occasion'),
    )



from onegov.core.orm import Base
from onegov.core.orm.mixins import TimestampMixin
from onegov.core.orm.types import UUID
from sqlalchemy import Column, ForeignKey
from sqlalchemy.orm import relationship
from uuid import uuid4


class PublicationRequest(Base, TimestampMixin):
    """ Describes a request for publication. As users create new activities
    they need to ask for publication, where the activity is reviewed before
    it is made public.

    This repeats each period. Every activity which should be provided again
    has to be requested again. This is even possible multiple times per
    period, though that should be the exception.

    """

    __tablename__ = 'publication_requests'

    #: The public id of the publication request
    id = Column(UUID, primary_key=True, default=uuid4)

    #: The activity linked to this request
    activity_id = Column(UUID, ForeignKey('activities.id'), nullable=False)
    activity = relationship(
        'Activity', backref='publication_requests', lazy='joined')

    #: The period linked to this request
    period_id = Column(UUID, ForeignKey("periods.id"), nullable=False)
    period = relationship(
        'Period', backref='publication_requests', lazy='joined')



import sedate

from datetime import time
from enum import IntEnum
from onegov.core.orm import Base
from onegov.core.orm.mixins import TimestampMixin
from onegov.core.orm.types import UUID, UTCDateTime
from sqlalchemy import event
from sqlalchemy import CheckConstraint
from sqlalchemy import Column
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import Text
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import Session


class DAYS(IntEnum):
    half = 2**1
    full = 2**2
    many = 2**3

    @staticmethod
    def has(value, mask):
        return value & mask > 0 if value else False

    @staticmethod
    def compute(localized_start, localized_end, total_seconds):
        hours = total_seconds / 3600

        if hours <= 6:
            return DAYS.half
        elif hours <= 24:
            start, end = localized_start, localized_end

            # if a less than 24 hours long activity ends on another day, the
            # end time is relevant. An end before 06:00 indicates that this
            # is an activity that lasts very long. An end after 06:00 is an
            # multi-day activity.
            if start.date() != end.date() and end.time() >= time(6, 0):
                return DAYS.many

            return DAYS.full
        else:
            return DAYS.many


class OccasionDate(Base, TimestampMixin):
    """ A single start/end date of an occurrence (which may have multiple
    date ranges).

    """

    __tablename__ = 'occasion_dates'

    def __hash__(self):
        return hash(self.id)

    #: the internal id of this occasion date
    id = Column(Integer, primary_key=True)

    #: Timezone of the occasion date
    timezone = Column(Text, nullable=False)

    #: The start of the range
    start = Column(UTCDateTime, nullable=False)

    #: The end of the range
    end = Column(UTCDateTime, nullable=False)

    #: The associated occasion
    occasion_id = Column(UUID, ForeignKey('occasions.id'), nullable=False)

    __table_args__ = (
        CheckConstraint('"start" <= "end"', name='start_before_end'),
    )

    @property
    def localized_start(self):
        return sedate.to_timezone(self.start, self.timezone)

    @property
    def localized_end(self):
        return sedate.to_timezone(self.end, self.timezone)

    @property
    def active_days(self):
        for dt in sedate.dtrange(self.localized_start, self.localized_end):
            yield dt.date().toordinal()

    @property
    def weekdays(self):
        return list({
            dt.weekday() for dt in sedate.dtrange(
                self.localized_start, self.localized_end
            )
        })

    @hybrid_property
    def duration_in_seconds(self):
        return (self.end - self.start).total_seconds()

    @hybrid_property
    def duration(self):
        return DAYS.compute(
            self.localized_start,
            self.localized_end,
            self.duration_in_seconds)

    def overlaps(self, other):
        return sedate.overlaps(self.start, self.end, other.start, other.end)


# # changes to the dates need to be propagated to the parent occasion
# # so it can update its aggreagated values
@event.listens_for(Session, 'before_flush')
def before_flush(session, context, instances):
    for obj in session.dirty:
        if isinstance(obj, OccasionDate):
            obj.occasion.on_date_change()



from onegov.activity.utils import merge_ranges
from onegov.activity.utils import extract_municipality


def test_merge_ranges():
    assert merge_ranges([[1, 2]]) == [(1, 2)]
    assert merge_ranges([[1, 2], [2, 3]]) == [(1, 3)]
    assert merge_ranges([[1, 2], [2, 3]]) == [(1, 3)]
    assert merge_ranges([[1, 2], [3, 4]]) == [(1, 2), (3, 4)]
    assert merge_ranges([[1, 2], [1, 2], [2, 3]]) == [(1, 3)]


def test_extract_municipality():
    assert extract_municipality("6004 Luzern") == (6004, 'Luzern')
    assert extract_municipality("9000 St. Gallen") == (9000, 'St. Gallen')
    assert extract_municipality("""
        Bahnhofstrasse 123
        1234 Test
    """) == (1234, "Test")
    assert extract_municipality("Bahnhofstrasse 123, 1234 Test")\
        == (1234, "Test")

    assert extract_municipality("""
        1234 Foo
        3456 Bar
    """) == (1234, "Foo")

    assert extract_municipality("0123 invalid plz") is None
    assert extract_municipality("4653 Obergösgen") == (4653, "Obergösgen")



from onegov.activity.models import Period
from onegov.core.collection import GenericCollection


class PeriodCollection(GenericCollection):

    @property
    def model_class(self):
        return Period

    def add(self, title, prebooking, execution, active=False,
            minutes_between=0, deadline_date=None, deadline_days=None,
            cancellation_date=None, cancellation_days=None):

        return super().add(
            title=title,
            prebooking_start=prebooking[0],
            prebooking_end=prebooking[1],
            execution_start=execution[0],
            execution_end=execution[1],
            minutes_between=minutes_between,
            active=active,
            deadline_date=deadline_date,
            deadline_days=deadline_days,
            cancellation_date=cancellation_date,
            cancellation_days=cancellation_days,
        )

    def active(self):
        return self.query().filter(Period.active == True).first()



import lxml
import random
import re
import sedate
import string

from datetime import date, timedelta
from functools import partial
from pyquery import PyQuery as pq

INTERNAL_IMAGE_EX = re.compile(r'.*/storage/[0-9a-z]{64}')

NUM_RANGE_RE = re.compile(r'\d+-\d+')
DATE_RANGE_RE = re.compile(r'\d{4}-\d{2}-\d{2}:\d{4}-\d{2}-\d{2}')

MUNICIPALITY_EX = re.compile(r"""
    (?P<zipcode>[1-9]{1}[0-9]{3})
    \s+
    (?P<municipality>[\w\s\(\)\.\-]+)
""", re.VERBOSE)


GROUP_CODE_EX = re.compile(r'[A-Z]{3}-?[A-Z]{3}-?[A-Z]{3}')


def random_group_code():
    # 26^9 should be a decent amount of codes to randomly chose, without
    # having to check their uniqueness
    raw = ''.join(random.choice(string.ascii_uppercase) for x in range(9))

    return '-'.join((raw[:3], raw[3:6], raw[-3:]))


def is_valid_group_code(code):
    return GROUP_CODE_EX.match(code) and True or False


def overlaps(range_a, range_b):
    return range_b[0] <= range_a[0] and range_a[0] <= range_b[1] or\
        range_a[0] <= range_b[0] and range_b[0] <= range_a[1]


def merge_ranges(ranges):
    """ Merges the given list of ranges into a list of ranges including only
    exclusive ranges. The ranges are turned into tuples to make them
    hashable.

    """

    ranges = sorted(list(ranges))

    # stack of merged values
    merged = [tuple(ranges[0])]

    for r in ranges:
        if overlaps(merged[-1], r):
            merged[-1] = (merged[-1][0], r[1])
        else:
            merged.append(tuple(r))

    return merged


def num_range_decode(s):
    if not isinstance(s, str):
        return None

    if not NUM_RANGE_RE.match(s):
        return None

    age_range = tuple(int(a) for a in s.split('-'))

    if age_range[0] <= age_range[1]:
        return age_range
    else:
        return None


def num_range_encode(a):
    return '-'.join(str(n) for n in a)


def date_range_decode(s):
    if not isinstance(s, str):
        return None

    if not DATE_RANGE_RE.match(s):
        return None

    s, e = s.split(':')

    return (
        date(*tuple(int(p) for p in s.split('-'))),
        date(*tuple(int(p) for p in e.split('-')))
    )


def date_range_encode(d):
    return ':'.join((d[0].strftime('%Y-%m-%d'), d[1].strftime('%Y-%m-%d')))


def generate_xml(payments):
    """ Creates an xml for import through ISO20022. Used for testing only. """

    transactions = []

    default = {
        'reference': '',
        'note': ''
    }

    for ix, payment in enumerate(payments):

        if 'tid' not in payment:
            payment['tid'] = 'T{}'.format(ix)

        if payment['amount'].startswith('-'):
            payment['credit'] = 'DBIT'
        else:
            payment['credit'] = 'CRDT'

        payment['currency'] = payment['amount'][-3:]
        payment['amount'] = payment['amount'].strip('-+')[:-3]

        for key in default:
            if key not in payment:
                payment[key] = default[key]

        transactions.append("""
        <TxDtls>
            <Refs>
                <AcctSvcrRef>{tid}</AcctSvcrRef>
            </Refs>
            <Amt Ccy="{currency}">{amount}</Amt>
            <CdtDbtInd>{credit}</CdtDbtInd>
            <RmtInf>
                <Strd>
                    <CdtrRefInf>
                        <Ref>{reference}</Ref>
                    </CdtrRefInf>
                </Strd>
                <Ustrd>{note}</Ustrd>
            </RmtInf>
        </TxDtls>
        """.format(**payment))

    return """<?xml version="1.0" encoding="UTF-8"?>
        <Document>
            <BkToCstmrStmt>
                <Stmt>
                    <Ntry>
                        <NtryDtls>
                            {}
                        </NtryDtls>
                    </Ntry>
                </Stmt>
            </BkToCstmrStmt>
        </Document>
    """.format('\n'.join(transactions))


def dates_overlap(a, b, minutes_between=0, cut_end=True, alignment=None):
    """ Returns true if any time tuple in the list of tuples in a overlaps
    with a time tuple in b.

    """
    # this can be done with an O(n log n) algorithm but since we are
    # operating on a very small n the constant factors dominate and there
    # are fewer constant factors in this approach:

    offset = timedelta(seconds=minutes_between / 2 * 60)
    ms = cut_end and timedelta(microseconds=1) or timedelta()

    # make sure that 11:00 - 12:00 and 12:00 - 13:00 are not overlapping
    ms = timedelta(microseconds=1)

    if alignment:

        # alignment function
        align = getattr(sedate, f'align_range_to_{alignment}')

        # it is highliy unlikely that this will ever be anything else as this
        # module is pretty much tailored for Switzerland
        align = partial(align, timezone='Europe/Zurich')

    for s, e in a:
        for os, oe in b:

            if alignment:
                s, e = align(s, e)
                os, oe = align(os, oe)

            if sedate.overlaps(
                s - offset, e + offset - ms,
                os - offset, oe + offset - ms
            ):
                return True

    return False


def is_internal_image(url):
    return url and INTERNAL_IMAGE_EX.match(url) and True or False


def extract_thumbnail(text):

    try:
        first_image = next((img for img in pq(text or '')('img')), None)
    except (lxml.etree.XMLSyntaxError, lxml.etree.ParserError):
        first_image = None

    if first_image is None:
        return None

    url = first_image.attrib.get('src')

    if is_internal_image(url) and not url.endswith('/thumbnail'):
        url += '/thumbnail'

    return url


def extract_municipality(text):
    for line in text.splitlines():
        for fragment in line.split(','):
            match = MUNICIPALITY_EX.match(fragment.strip())

            if match:
                return int(match.group('zipcode')), match.group('municipality')



from onegov.activity.models.invoice_item import InvoiceItem, SCALE
from onegov.activity.models.period import Period
from onegov.core.orm import Base
from onegov.core.orm.mixins import TimestampMixin
from onegov.core.orm.types import UUID
from onegov.pay import Price
from onegov.user import User
from sqlalchemy import and_
from sqlalchemy import Column
from sqlalchemy import ForeignKey
from sqlalchemy import func
from sqlalchemy import select
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import object_session, relationship
from uuid import uuid4


class Invoice(Base, TimestampMixin):
    """ A grouping of invoice items. """

    __tablename__ = 'invoices'

    #: the public id of the invoice
    id = Column(UUID, primary_key=True, default=uuid4)

    #: the period to which this invoice belongs to
    period_id = Column(UUID, ForeignKey('periods.id'), nullable=False)
    period = relationship(Period, backref='invoices')

    #: the user to which the invoice belongs
    user_id = Column(UUID, ForeignKey('users.id'), nullable=False)
    user = relationship(User, backref='invoices')

    #: the specific items linked with this invoice
    items = relationship(InvoiceItem, backref='invoice')

    @property
    def price(self):
        return Price(self.outstanding_amount, 'CHF')

    def readable_by_bucket(self, bucket):
        for ref in self.references:
            if ref.bucket == bucket:
                return ref.readable

        return None

    def sync(self):
        items = object_session(self).query(InvoiceItem).filter(and_(
            InvoiceItem.source != None,
            InvoiceItem.source != 'xml'
        )).join(InvoiceItem.payments)

        for item in (i for i in items if i.payments):
            for payment in item.payments:

                # though it should be fairly rare, it's possible for
                # charges not to be captured yet
                if payment.state == 'open':
                    payment.charge.capture()
                    payment.sync()

            # the last payment is the relevant one
            item.paid = item.payments[-1].state == 'paid'

    def add(self, group, text, unit, quantity, flush=True, **kwargs):
        item = InvoiceItem(
            group=group,
            text=text,
            unit=unit,
            quantity=quantity,
            invoice_id=self.id,
            **kwargs
        )

        self.items.append(item)

        if flush:
            object_session(self).flush()

        return item

    @hybrid_property
    def discourage_changes(self):
        for item in self.items:
            if item.source == 'xml':
                return True

        return False

    @hybrid_property
    def disable_changes(self):
        for item in self.items:
            if not item.source:
                continue

            if item.source == 'xml':
                continue

            states = {p.state for p in item.payments}

            if 'open' in states or 'paid' in states:
                return True

    @hybrid_property
    def has_online_payments(self):
        for item in self.items:
            if not item.source or item.source == 'xml':
                continue

            True

    # paid or not
    @hybrid_property
    def paid(self):
        return self.outstanding_amount <= 0

    # paid + unpaid
    @hybrid_property
    def total_amount(self):
        return self.outstanding_amount + self.paid_amount

    @total_amount.expression
    def total_amount(cls):
        return select([func.sum(InvoiceItem.amount)]).\
            where(InvoiceItem.invoice_id == cls.id).\
            label('total_amount')

    # paid only
    @hybrid_property
    def outstanding_amount(self):
        return round(
            sum(item.amount for item in self.items if not item.paid),
            SCALE
        )

    @outstanding_amount.expression
    def outstanding_amount(cls):
        return select([func.sum(InvoiceItem.amount)]).\
            where(and_(
                InvoiceItem.invoice_id == cls.id,
                InvoiceItem.paid == False
            )).label('outstanding_amount')

    # unpaid only
    @hybrid_property
    def paid_amount(self):
        return round(
            sum(item.amount for item in self.items if item.paid),
            SCALE
        )

    @paid_amount.expression
    def paid_amount(cls):
        return select([func.sum(InvoiceItem.amount)]).\
            where(and_(
                InvoiceItem.invoice_id == cls.id,
                InvoiceItem.paid == True
            )).label('paid_amount')
